# Based on lxd.py (c) 2016, Matt Clay <matt@mystile.com>
# (c) 2023, Stephane Graber <stgraber@stgraber.org>
# Copyright (c) 2023 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author: Stéphane Graber (@stgraber)
name: incus
short_description: Run tasks in Incus instances using the Incus CLI
description:
  - Run commands or put/fetch files to an existing Incus instance using Incus CLI.
version_added: "8.2.0"
notes:
  - When using this collection for Windows virtual machines, set C(ansible_shell_type) to C(powershell) or C(cmd) as a variable to the host in
    the inventory.
options:
  remote_addr:
    description:
      - The instance identifier.
    type: string
    default: inventory_hostname
    vars:
      - name: inventory_hostname
      - name: ansible_host
      - name: ansible_incus_host
  executable:
    description:
      - The shell to use for execution inside the instance.
    type: string
    default: /bin/sh
    vars:
      - name: ansible_executable
      - name: ansible_incus_executable
  incus_become_method:
    description:
      - Become command used to switch to a non-root user.
      - Is only used when O(remote_user) is not V(root).
    type: str
    default: /bin/su
    vars:
      - name: incus_become_method
    version_added: 10.4.0
  remote:
    description:
      - The name of the Incus remote to use (per C(incus remote list)).
      - Remotes are used to access multiple servers from a single client.
    type: string
    default: local
    vars:
      - name: ansible_incus_remote
  remote_user:
    description:
      - User to login/authenticate as.
      - Can be set from the CLI with the C(--user) or C(-u) options.
    type: string
    default: root
    vars:
      - name: ansible_user
    env:
      - name: ANSIBLE_REMOTE_USER
    ini:
      - section: defaults
        key: remote_user
    keyword:
      - name: remote_user
    version_added: 10.4.0
  project:
    description:
      - The name of the Incus project to use (per C(incus project list)).
      - Projects are used to divide the instances running on a server.
    type: string
    default: default
    vars:
      - name: ansible_incus_project
"""

import os
import re
import shlex
from subprocess import PIPE, Popen, call

from ansible.errors import AnsibleConnectionFailure, AnsibleError, AnsibleFileNotFound
from ansible.module_utils.common.process import get_bin_path
from ansible.module_utils.common.text.converters import to_bytes
from ansible.plugins.connection import ConnectionBase


class Connection(ConnectionBase):
    """Incus based connections"""

    transport = "incus"
    has_pipelining = True

    def __init__(self, play_context, new_stdin, *args, **kwargs):
        super().__init__(play_context, new_stdin, *args, **kwargs)

        self._incus_cmd = get_bin_path("incus")

        if not self._incus_cmd:
            raise AnsibleError("incus command not found in PATH")

        if getattr(self._shell, "_IS_WINDOWS", False):
            # Initializing regular expression patterns to match on a PowerShell or cmd command line.
            self.powershell_regex_pattern = re.compile(
                r'^"?(?P<executable>(?:[a-z]:\\)?[a-z0-9 ()\\.]*powershell(?:\.exe)?)"?\s+(?P<args>.*)(?P<command>-c(?:ommand)?)\s+(?P<post_args>.*(\n.*)*)',
                re.IGNORECASE,
            )
            self.cmd_regex_pattern = re.compile(
                r'^"?(?P<executable>(?:[a-z]:\\)?[a-z0-9 ()\\.]*cmd(?:\.exe)?)"?\s+(?P<args>.*)(?P<command>/c)\s+(?P<post_args>.*)',
                re.IGNORECASE,
            )

            # Basic setup for a Windows host.
            self.has_native_async = True
            self.always_pipeline_modules = True
            self.module_implementation_preferences = (".ps1", ".exe", "")
            self.allow_executable = False

    def _connect(self):
        """connect to Incus (nothing to do here)"""
        super()._connect()

        if not self._connected:
            self._display.vvv(
                f"ESTABLISH Incus CONNECTION FOR USER: {self.get_option('remote_user')}", host=self._instance()
            )
            self._connected = True

    def _build_command(self, cmd) -> list[str]:
        """build the command to execute on the incus host"""

        # Force pseudo-terminal allocation if the active become plugin
        # requires one (e.g. community.general.machinectl), otherwise the
        # become helper runs without a controlling tty and silently fails.
        require_tty = self.become is not None and getattr(self.become, "require_tty", False)

        exec_cmd: list[str] = [
            self._incus_cmd,
            "--project",
            self.get_option("project"),
            "exec",
            *(["-T"] if getattr(self._shell, "_IS_WINDOWS", False) else []),
            *(["-t"] if require_tty and not getattr(self._shell, "_IS_WINDOWS", False) else []),
            f"{self.get_option('remote')}:{self._instance()}",
            "--",
        ]

        if getattr(self._shell, "_IS_WINDOWS", False):
            if regex_match := self.powershell_regex_pattern.match(cmd):
                regex_pattern = self.powershell_regex_pattern
            elif regex_match := self.cmd_regex_pattern.match(cmd):
                regex_pattern = self.cmd_regex_pattern

            if regex_match:
                self._display.vvvvvv(
                    f'Found keyword: "{regex_match.group("command")}" based on regex: {regex_pattern.pattern}',
                    host=self._instance(),
                )

                # To avoid splitting on a space contained in the path, set the executable as the first argument.
                exec_cmd.append(regex_match.group("executable"))
                if args := regex_match.group("args"):
                    exec_cmd.extend(args.strip().split(" "))

                # Set the command argument depending on cmd or powershell and the rest of it
                exec_cmd.append(regex_match.group("command"))

                if post_args := regex_match.group("post_args"):
                    post_args = post_args.strip()

                    # Keep quotes from becoming literal PowerShell argument content.
                    if len(post_args) >= 2 and post_args[0] == post_args[-1] and post_args[0] in ("'", '"'):
                        self._display.v(
                            "WARNING: PowerShell -Command argument is wrapped in outer quotes; "
                            "this connection plugin strips those quotes and behavior may differ "
                            "from a direct shell run. Prefer passing -Command without extra "
                            "outer quoting.",
                            host=self._instance(),
                        )
                        post_args = post_args[1:-1]

                    exec_cmd.append(post_args)
            else:
                # Keep quotes from becoming literal PowerShell argument content.
                # shlex is not a full PowerShell/Windows command-line parser,
                # but with posix=False it reliably keeps quoted segments (for
                # example, paths with spaces) as single argv items, which is
                # what we need before passing arguments to incus exec.
                parts = shlex.split(cmd, posix=False)
                for i, part in enumerate(parts[:-1]):
                    if part.lower() in {"-enc", "-encodedcommand", "-file", "-f"}:
                        parts[i + 1] = parts[i + 1].strip("'\"")
                        break
                exec_cmd.extend(parts)
        else:
            if self.get_option("remote_user") != "root":
                self._display.vvv(
                    f"INFO: Running as non-root user: {self.get_option('remote_user')}, \
                  trying to run 'incus exec' with become method: {self.get_option('incus_become_method')}",
                    host=self._instance(),
                )
                exec_cmd.extend([self.get_option("incus_become_method"), self.get_option("remote_user"), "-c"])

            exec_cmd.extend([self.get_option("executable"), "-c", cmd])

        return exec_cmd

    def _instance(self):
        # Return only the leading part of the FQDN as the instance name
        # as Incus instance names cannot be a FQDN.
        return self.get_option("remote_addr").split(".")[0]

    def exec_command(self, cmd, in_data=None, sudoable=True):
        """execute a command on the Incus host"""
        super().exec_command(cmd, in_data=in_data, sudoable=sudoable)

        self._display.vvv(f"EXEC {cmd}", host=self._instance())

        local_cmd = self._build_command(cmd)
        self._display.vvvvv(f"EXEC {local_cmd}", host=self._instance())

        local_cmd = [to_bytes(i, errors="surrogate_or_strict") for i in local_cmd]
        in_data = to_bytes(in_data, errors="surrogate_or_strict", nonstring="passthru")

        process = Popen(local_cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate(in_data)

        if stderr.startswith(b"Error: ") and stderr.rstrip().endswith(b": Instance is not running"):
            raise AnsibleConnectionFailure(
                f"instance not running: {self._instance()} (remote={self.get_option('remote')}, project={self.get_option('project')})"
            )

        if stderr.startswith(b"Error: ") and stderr.rstrip().endswith(b": Instance not found"):
            raise AnsibleConnectionFailure(
                f"instance not found: {self._instance()} (remote={self.get_option('remote')}, project={self.get_option('project')})"
            )

        if stderr.startswith(b"Error: ") and b": User does not have permission " in stderr:
            raise AnsibleConnectionFailure(
                f"instance access denied: {self._instance()} (remote={self.get_option('remote')}, project={self.get_option('project')})"
            )

        if stderr.startswith(b"Error: ") and b": User does not have entitlement " in stderr:
            raise AnsibleConnectionFailure(
                f"instance access denied: {self._instance()} (remote={self.get_option('remote')}, project={self.get_option('project')})"
            )

        return process.returncode, stdout, stderr

    def _get_remote_uid_gid(self) -> tuple[int, int]:
        """Get the user and group ID of 'remote_user' from the instance."""

        rc, uid_out, err = self.exec_command("/bin/id -u")
        if rc != 0:
            raise AnsibleError(f"Failed to get remote uid for user {self.get_option('remote_user')}: {err}")
        uid = uid_out.strip()

        rc, gid_out, err = self.exec_command("/bin/id -g")
        if rc != 0:
            raise AnsibleError(f"Failed to get remote gid for user {self.get_option('remote_user')}: {err}")
        gid = gid_out.strip()

        return int(uid), int(gid)

    def put_file(self, in_path, out_path):
        """put a file from local to Incus"""
        super().put_file(in_path, out_path)

        self._display.vvv(f"PUT {in_path} TO {out_path}", host=self._instance())

        if not os.path.isfile(to_bytes(in_path, errors="surrogate_or_strict")):
            raise AnsibleFileNotFound(f"input path is not a file: {in_path}")

        if not getattr(self._shell, "_IS_WINDOWS", False) and self.get_option("remote_user") != "root":
            uid, gid = self._get_remote_uid_gid()
            local_cmd = [
                self._incus_cmd,
                "--project",
                self.get_option("project"),
                "file",
                "push",
                "--uid",
                str(uid),
                "--gid",
                str(gid),
                "--quiet",
                in_path,
                f"{self.get_option('remote')}:{self._instance()}/{out_path}",
            ]
        else:
            local_cmd = [
                self._incus_cmd,
                "--project",
                self.get_option("project"),
                "file",
                "push",
                "--quiet",
                in_path,
                f"{self.get_option('remote')}:{self._instance()}/{out_path}",
            ]

        self._display.vvvvv(f"PUT {local_cmd}", host=self._instance())

        local_cmd = [to_bytes(i, errors="surrogate_or_strict") for i in local_cmd]

        call(local_cmd)

    def fetch_file(self, in_path, out_path):
        """fetch a file from Incus to local"""
        super().fetch_file(in_path, out_path)

        self._display.vvv(f"FETCH {in_path} TO {out_path}", host=self._instance())

        local_cmd = [
            self._incus_cmd,
            "--project",
            self.get_option("project"),
            "file",
            "pull",
            "--quiet",
            f"{self.get_option('remote')}:{self._instance()}/{in_path}",
            out_path,
        ]

        local_cmd = [to_bytes(i, errors="surrogate_or_strict") for i in local_cmd]

        call(local_cmd)

    def close(self):
        """close the connection (nothing to do here)"""
        super().close()

        self._connected = False



# Copyright (c) 2016 Matt Davis, <mdavis@ansible.com>
#                    Chris Houseknecht, <house@redhat.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


import os
import re
import types
import copy
import inspect
import traceback
import json

from os.path import expanduser

from ansible.module_utils.basic import \
    AnsibleModule, missing_required_lib, env_fallback

try:
    from ansible.module_utils.ansible_release import __version__ as ANSIBLE_VERSION
except Exception:
    ANSIBLE_VERSION = 'unknown'
from ansible.module_utils.six.moves import configparser
import ansible.module_utils.six.moves.urllib.parse as urlparse

AZURE_COMMON_ARGS = dict(
    auth_source=dict(
        type='str',
        choices=['auto', 'cli', 'env', 'credential_file', 'msi'],
        fallback=(env_fallback, ['ANSIBLE_AZURE_AUTH_SOURCE']),
        default="auto"
    ),
    profile=dict(type='str'),
    subscription_id=dict(type='str'),
    client_id=dict(type='str', no_log=True),
    secret=dict(type='str', no_log=True),
    tenant=dict(type='str', no_log=True),
    ad_user=dict(type='str', no_log=True),
    password=dict(type='str', no_log=True),
    cloud_environment=dict(type='str', default='AzureCloud'),
    cert_validation_mode=dict(type='str', choices=['validate', 'ignore']),
    api_profile=dict(type='str', default='latest'),
    adfs_authority_url=dict(type='str', default=None),
    log_mode=dict(type='str', no_log=True),
    log_path=dict(type='str', no_log=True),
    x509_certificate_path=dict(type='path', no_log=True),
    thumbprint=dict(type='str', no_log=True),
    disable_instance_discovery=dict(type='bool', default=False),
    oidc_token=dict(type='str', no_log=True),  # direct JWT assertion
    oidc_token_file_path=dict(type='path', no_log=True)  # path to JWT assertion file
)

AZURE_CREDENTIAL_ENV_MAPPING = dict(
    profile='AZURE_PROFILE',
    subscription_id='AZURE_SUBSCRIPTION_ID',
    client_id='AZURE_CLIENT_ID',
    secret='AZURE_SECRET',
    tenant='AZURE_TENANT',
    ad_user='AZURE_AD_USER',
    password='AZURE_PASSWORD',
    cloud_environment='AZURE_CLOUD_ENVIRONMENT',
    cert_validation_mode='AZURE_CERT_VALIDATION_MODE',
    adfs_authority_url='AZURE_ADFS_AUTHORITY_URL',
    x509_certificate_path='AZURE_X509_CERTIFICATE_PATH',
    thumbprint='AZURE_THUMBPRINT',
    disable_instance_discovery='AZURE_DISABLE_INSTANCE_DISCOVERY',
    oidc_token='AZURE_FEDERATED_TOKEN',
    oidc_token_file_path='AZURE_FEDERATED_TOKEN_FILE'
)


class SDKProfile(object):  # pylint: disable=too-few-public-methods

    def __init__(self, default_api_version, profile=None):
        """Constructor.

        :param str default_api_version: Default API version if not overridden by a profile. Nullable.
        :param profile: A dict operation group name to API version.
        :type profile: dict[str, str]
        """
        self.profile = profile if profile is not None else {}
        self.profile[None] = default_api_version

    @property
    def default_api_version(self):
        return self.profile[None]


# FUTURE: this should come from the SDK or an external location.
# For now, we have to copy from azure-cli
AZURE_API_PROFILES = {
    'latest': {
        'AuthorizationManagementClient': '2020-04-01-preview',
        'ContainerInstanceManagementClient': '2018-02-01-preview',
        'ComputeManagementClient': dict(
            default_api_version='2018-10-01',
            resource_skus='2018-10-01',
            disks='2018-06-01',
            snapshots='2018-10-01',
            virtual_machine_run_commands='2018-10-01'
        ),
        'ManagementGroupsClient': '2020-05-01',
        'NetworkManagementClient': '2019-11-01',
        'ResourceManagementClient': '2019-10-01',
        'SearchManagementClient': '2020-08-01',
        'StorageManagementClient': '2021-06-01',
        'SubscriptionClient': '2019-11-01',
        'WebSiteManagementClient': '2021-03-01',
        'PostgreSQLManagementClient': '2017-12-01',
        'MySQLManagementClient': '2017-12-01',
        'MariaDBManagementClient': '2019-03-01',
        'ManagementLockClient': '2016-09-01',
        'DataLakeStoreAccountManagementClient': '2016-11-01',
        'NotificationHubsManagementClient': '2016-03-01',
        'EventHubManagementClient': '2021-11-01',
        'GenericRestClient': 'latest',
        'DnsManagementClient': '2018-05-01',
        'PrivateDnsManagementClient': 'latest',
        'ContainerServiceClient': '2022-02-01',
        'SqlManagementClient': 'latest',
        'ContainerRegistryManagementClient': '2021-09-01',
        'MarketplaceOrderingAgreements': 'latest',
        'TrafficManagerManagementClient': 'latest',
        'MonitorManagementClient': '2016-03-01',
        'LogAnalyticsManagementClient': 'latest',
        'ServiceBusManagementClient': 'latest',
        'AutomationClient': 'latest',
        'IotHubClient': 'latest',
        'RecoveryServicesBackupClient': 'latest',
        'DataFactoryManagementClient': 'latest',
        'KeyVaultManagementClient': '2026-02-01',
        'HDInsightManagementClient': 'latest',
        'DevTestLabsClient': 'latest',
        'CosmosDBManagementClient': 'latest',
        'CdnManagementClient': '2017-04-02',
        'BatchManagementClient': 'latest',
        'EventGridManagementClient': '2025-02-15',
        'AppConfigurationManagementClient': '2024-05-01',
        'HybridComputeManagementClient': 'latest'
    },
    '2019-03-01-hybrid': {
        'StorageManagementClient': '2017-10-01',
        'NetworkManagementClient': '2017-10-01',
        'ComputeManagementClient': SDKProfile('2017-12-01', {
            'resource_skus': '2017-09-01',
            'disks': '2017-03-30',
            'snapshots': '2017-03-30'
        }),
        'ManagementLinkClient': '2016-09-01',
        'ManagementLockClient': '2016-09-01',
        'PolicyClient': '2016-12-01',
        'ResourceManagementClient': '2018-05-01',
        'EventHubManagementClient': '2018-05-04',
        'SubscriptionClient': '2016-06-01',
        'DnsManagementClient': '2016-04-01',
        'KeyVaultManagementClient': '2016-10-01',
        'AuthorizationManagementClient': SDKProfile('2015-07-01', {
            'classic_administrators': '2015-06-01',
            'policy_assignments': '2016-12-01',
            'policy_definitions': '2016-12-01'
        }),
        'KeyVaultClient': '2016-10-01',
        'azure.multiapi.storage': '2017-11-09',
        'azure.multiapi.cosmosdb': '2017-04-17'
    },
    '2018-03-01-hybrid': {
        'StorageManagementClient': '2016-01-01',
        'NetworkManagementClient': '2017-10-01',
        'ComputeManagementClient': SDKProfile('2017-03-30'),
        'ManagementLinkClient': '2016-09-01',
        'ManagementLockClient': '2016-09-01',
        'PolicyClient': '2016-12-01',
        'ResourceManagementClient': '2018-02-01',
        'SubscriptionClient': '2016-06-01',
        'DnsManagementClient': '2016-04-01',
        'KeyVaultManagementClient': '2016-10-01',
        'AuthorizationManagementClient': SDKProfile('2015-07-01', {
            'classic_administrators': '2015-06-01'
        }),
        'KeyVaultClient': '2016-10-01',
        'azure.multiapi.storage': '2017-04-17',
        'azure.multiapi.cosmosdb': '2017-04-17'
    },
    '2017-03-09-profile': {
        'StorageManagementClient': '2016-01-01',
        'NetworkManagementClient': '2015-06-15',
        'ComputeManagementClient': SDKProfile('2016-03-30'),
        'ManagementLinkClient': '2016-09-01',
        'ManagementLockClient': '2015-01-01',
        'PolicyClient': '2015-10-01-preview',
        'ResourceManagementClient': '2016-02-01',
        'SubscriptionClient': '2016-06-01',
        'DnsManagementClient': '2016-04-01',
        'KeyVaultManagementClient': '2016-10-01',
        'AuthorizationManagementClient': SDKProfile('2015-07-01', {
            'classic_administrators': '2015-06-01'
        }),
        'KeyVaultClient': '2016-10-01',
        'azure.multiapi.storage': '2015-04-05'
    }
}

AZURE_TAG_ARGS = dict(
    tags=dict(type='dict'),
    append_tags=dict(type='bool', default=True),
)

AZURE_COMMON_REQUIRED_IF = [
    ('log_mode', 'file', ['log_path'])
]

ANSIBLE_USER_AGENT = 'Ansible/{0}'.format(ANSIBLE_VERSION)
CLOUDSHELL_USER_AGENT_KEY = 'AZURE_HTTP_USER_AGENT'
VSCODEEXT_USER_AGENT_KEY = 'VSCODEEXT_USER_AGENT'

CIDR_PATTERN = re.compile(r"(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1"
                          r"[0-9]{2}|2[0-4][0-9]|25[0-5])(/([0-9]|[1-2][0-9]|3[0-2]))")

AZURE_SUCCESS_STATE = "Succeeded"

AZURE_IMPORT_ERROR = None


try:
    import importlib
except ImportError:
    # This passes the sanity import test, but does not provide a user friendly error message.
    # Doing so would require catching Exception for all imports of Azure dependencies in modules and module_utils.
    importlib = None

# Imports
try:
    from enum import Enum
    from azure.mgmt.core.tools import parse_resource_id, resource_id, is_valid_resource_id
    from azure.mgmt.network import NetworkManagementClient
    from azure.mgmt.network import models as NetworkModels
    from azure.mgmt.resource.resources import ResourceManagementClient
    from azure.mgmt.managementgroups import ManagementGroupsAPI as ManagementGroupsClient
    from azure.mgmt.resource.subscriptions import SubscriptionClient
    from azure.mgmt.storage import StorageManagementClient
    import azure.mgmt.storage.models as StorageModels
    from azure.mgmt.compute import ComputeManagementClient
    from azure.mgmt.dns import DnsManagementClient
    from azure.mgmt.privatedns import PrivateDnsManagementClient
    import azure.mgmt.privatedns.models as PrivateDnsModels
    from azure.mgmt.monitor import MonitorManagementClient
    from azure.mgmt.web import WebSiteManagementClient
    from azure.mgmt.containerservice import ContainerServiceClient
    from azure.mgmt.marketplaceordering import MarketplaceOrderingAgreements
    from azure.mgmt.trafficmanager import TrafficManagerManagementClient
    from azure.storage.blob import BlobServiceClient
    from azure.storage.fileshare import ShareDirectoryClient, ShareClient, ShareFileClient
    from azure.mgmt.authorization import AuthorizationManagementClient
    from azure.mgmt.sql import SqlManagementClient
    from azure.mgmt.servicebus import ServiceBusManagementClient
    from azure.mgmt.rdbms.postgresql import PostgreSQLManagementClient
    from azure.mgmt.postgresqlflexibleservers import PostgreSQLManagementClient as PostgreSQLFlexibleManagementClient
    from azure.mgmt.rdbms.mysql import MySQLManagementClient
    from azure.mgmt.mysqlflexibleservers import MySQLManagementClient as MySQLFlexibleManagementClient
    from azure.mgmt.rdbms.mariadb import MariaDBManagementClient
    from azure.mgmt.containerregistry import ContainerRegistryManagementClient
    from azure.mgmt.containerinstance import ContainerInstanceManagementClient
    from azure.mgmt.loganalytics import LogAnalyticsManagementClient
    import azure.mgmt.loganalytics.models as LogAnalyticsModels
    from azure.mgmt.automation import AutomationClient
    import azure.mgmt.automation.models as AutomationModel
    from azure.mgmt.iothub import IotHubClient
    from azure.mgmt.iothub import models as IoTHubModels
    from azure.mgmt.resource.locks import ManagementLockClient
    from azure.mgmt.recoveryservicesbackup import RecoveryServicesBackupClient
    import azure.mgmt.recoveryservicesbackup.activestamp.models as RecoveryServicesBackupModels
    from azure.mgmt.search import SearchManagementClient
    from azure.mgmt.notificationhubs import NotificationHubsManagementClient
    from azure.mgmt.eventhub import EventHubManagementClient
    from azure.mgmt.datafactory import DataFactoryManagementClient
    import azure.mgmt.datafactory.models as DataFactoryModel
    from azure.identity._credentials import client_secret, user_password, certificate, managed_identity
    from azure.identity import AzureCliCredential, ClientAssertionCredential
    from kiota_authentication_azure.azure_identity_authentication_provider import AzureIdentityAuthenticationProvider
    from msgraph_core import GraphClientFactory, NationalClouds
    from msgraph import GraphRequestAdapter, GraphServiceClient
    from azure.mgmt.batch import BatchManagementClient
    from azure.mgmt.batch import models as BatchManagementModel
    from azure.mgmt.resourcehealth import ResourceHealthMgmtClient
    from azure.mgmt.cdn import CdnManagementClient
    from azure.mgmt.hybridcompute import HybridComputeManagementClient

except ImportError as exc:
    AZURE_IMPORT_ERROR = traceback.format_exc()

from base64 import b64encode, b64decode
from hashlib import sha256
from hmac import HMAC
from time import time
import subprocess

try:
    from urllib import (urlencode, quote_plus)
except ImportError:
    from urllib.parse import (urlencode, quote_plus)

try:
    from azure.cli.core.util import CLIError
    from azure.cli.core import cloud as azure_cloud
except ImportError:
    CLIError = Exception


def azure_id_to_dict(id):
    pieces = re.sub(r'^\/', '', id).split('/')
    result = {}
    index = 0
    while index < len(pieces) - 1:
        result[pieces[index]] = pieces[index + 1]
        index += 1
    return result


def format_resource_id(val, subscription_id, namespace, types, resource_group):
    return resource_id(name=val,
                       resource_group=resource_group,
                       namespace=namespace,
                       type=types,
                       subscription=subscription_id) if not is_valid_resource_id(val) else val


def normalize_location_name(name):
    return name.replace(' ', '').lower()


def _extract_missing_module(import_error):
    """
    Extract missing module name from ImportError traceback.
    Returns None if not detectable.
    """
    if not import_error:
        return None

    match = re.search(r"No module named '([^']+)'", import_error)
    if match:
        return match.group(1)

    return None


class AzureRMModuleBase(object):
    def __init__(self, derived_arg_spec, bypass_checks=False, no_log=False,
                 check_invalid_arguments=None, mutually_exclusive=None, required_together=None,
                 required_one_of=None, add_file_common_args=False, supports_check_mode=False,
                 required_if=None, supports_tags=True, facts_module=False, skip_exec=False,
                 is_ad_resource=False, required_by=None):

        merged_arg_spec = dict()
        merged_arg_spec.update(AZURE_COMMON_ARGS)
        if supports_tags:
            merged_arg_spec.update(AZURE_TAG_ARGS)

        if derived_arg_spec:
            merged_arg_spec.update(derived_arg_spec)

        merged_required_if = list(AZURE_COMMON_REQUIRED_IF)
        if required_if:
            merged_required_if += required_if

        if not required_by:
            required_by = dict()

        self.module = AnsibleModule(argument_spec=merged_arg_spec,
                                    bypass_checks=bypass_checks,
                                    no_log=no_log,
                                    mutually_exclusive=mutually_exclusive,
                                    required_together=required_together,
                                    required_one_of=required_one_of,
                                    add_file_common_args=add_file_common_args,
                                    supports_check_mode=supports_check_mode,
                                    required_if=merged_required_if,
                                    required_by=required_by)

        if AZURE_IMPORT_ERROR:
            missing_mod = _extract_missing_module(AZURE_IMPORT_ERROR)

            if missing_mod:
                msg = (
                    "Failed to import the required Python library ({0}). "
                    "Please install them by running: "
                    "pip install -r requirements.txt. "
                    "If the required library is installed, but Ansible is using the wrong "
                    "Python interpreter, please consult the documentation on "
                    "ansible_python_interpreter."
                ).format(missing_mod)
            else:
                msg = missing_required_lib("Azure SDK dependencies")

            self.fail(msg=msg, exception=AZURE_IMPORT_ERROR)

        self._authorization_client = None
        self._network_client = None
        self._storage_client = None
        self._subscription_client = None
        self._management_group_client = None
        self._resource_client = None
        self._compute_client = None
        self._disk_client = None
        self._image_version_client = None
        self._multi_disk_client = None
        self._diskencryptionset_client = None
        self._image_client = None
        self._dns_client = None
        self._private_dns_client = None
        self._web_client = None
        self._marketplace_client = None
        self._sql_client = None
        self._mysql_client = None
        self._mysql_flexible_client = None
        self._mariadb_client = None
        self._postgresql_client = None
        self._postgresql_flexible_client = None
        self._containerregistry_client = None
        self._containerregistrytoken_client = None
        self._containerinstance_client = None
        self._containerservice_client = None
        self._managedcluster_client = None
        self._traffic_manager_management_client = None
        self._monitor_autoscale_settings_client = None
        self._monitor_log_profiles_client = None
        self._monitor_diagnostic_settings_client = None
        self._monitor_data_collection_rules_client = None
        self._monitor_management_client_action_groups = None
        self._monitor_management_client_activity_log_alerts = None
        self._monitor_management_client_metric_alerts = None
        self._monitor_management_client_scheduled_query_rules = None
        self._resource = None
        self._log_analytics_client = None
        self._servicebus_client = None
        self._automation_client = None
        self._IoThub_client = None
        self._lock_client = None
        self._recovery_services_backup_client = None
        self._search_client = None
        self._datalake_store_client = None
        self._datafactory_client = None
        self._notification_hub_client = None
        self._event_hub_client = None
        self._batch_account_client = None
        self._resourcehealth_client = None
        self._cdn_client = None
        self._hybrid_compute_management_client = None

        self.check_mode = self.module.check_mode
        self.api_profile = self.module.params.get('api_profile')
        self.facts_module = facts_module
        # self.debug = self.module.params.get('debug')

        # delegate auth to AzureRMAuth class (shared with all plugin types)
        self.azure_auth = AzureRMAuth(fail_impl=self.fail, is_ad_resource=is_ad_resource, **self.module.params)

        # common parameter validation
        if self.module.params.get('tags'):
            self.validate_tags(self.module.params['tags'])

        if not skip_exec:
            res = self.exec_module(**self.module.params)
            self.module.exit_json(**res)

    def exec_module(self, **kwargs):
        self.fail("Error: {0} failed to implement exec_module method.".format(self.__class__.__name__))

    def fail(self, msg, **kwargs):
        '''
        Shortcut for calling module.fail()

        :param msg: Error message text.
        :param kwargs: Any key=value pairs
        :return: None
        '''
        self.module.fail_json(msg=msg, **kwargs)

    def deprecate(self, msg, version=None, collection_name='azure.azcollection'):
        self.module.deprecate(msg, version, collection_name=collection_name)

    def log(self, msg, pretty_print=False):
        if pretty_print:
            self.module.debug(json.dumps(msg, indent=4, sort_keys=True))
        else:
            self.module.debug(msg)

    def validate_tags(self, tags):
        '''
        Check if tags dictionary contains string:string pairs.

        :param tags: dictionary of string:string pairs
        :return: None
        '''
        if not self.facts_module:
            if not isinstance(tags, dict):
                self.fail("Tags must be a dictionary of string:string values.")
            for key, value in tags.items():
                if not isinstance(value, str):
                    self.fail("Tags values must be strings. Found {0}:{1}".format(str(key), str(value)))

    def update_tags(self, tags):
        '''
        Call from the module to update metadata tags. Returns tuple
        with bool indicating if there was a change and dict of new
        tags to assign to the object.

        :param tags: metadata tags from the object
        :return: bool, dict
        '''
        tags = tags or dict()
        new_tags = copy.copy(tags) if isinstance(tags, dict) else dict()
        param_tags = self.module.params.get('tags') if isinstance(self.module.params.get('tags'), dict) else dict()
        append_tags = self.module.params.get('append_tags') if self.module.params.get('append_tags') is not None else True
        changed = False
        # check add or update
        for key, value in param_tags.items():
            if not new_tags.get(key) or new_tags[key] != value:
                changed = True
                new_tags[key] = value
        # check remove
        if not append_tags:
            for key, value in tags.items():
                if not param_tags.get(key):
                    new_tags.pop(key)
                    changed = True
        return changed, new_tags

    def has_tags(self, obj_tags, tag_list):
        '''
        Used in fact modules to compare object tags to list of parameter tags. Return true if list of parameter tags
        exists in object tags.

        :param obj_tags: dictionary of tags from an Azure object.
        :param tag_list: list of tag keys or tag key:value pairs
        :return: bool
        '''

        if not obj_tags and tag_list:
            return False

        if not tag_list:
            return True

        matches = 0
        result = False
        for tag in tag_list:
            tag_key = tag
            tag_value = None
            if ':' in tag:
                tag_key, tag_value = tag.split(':')
            if tag_value and obj_tags.get(tag_key) == tag_value:
                matches += 1
            elif not tag_value and obj_tags.get(tag_key):
                matches += 1
        if matches == len(tag_list):
            result = True
        return result

    def get_resource_group(self, resource_group):
        '''
        Fetch a resource group.

        :param resource_group: name of a resource group
        :return: resource group object
        '''
        try:
            return self.rm_client.resource_groups.get(resource_group)
        except Exception as exc:
            self.fail("Error retrieving resource group {0} - {1}".format(resource_group, str(exc)))

    def parse_resource_to_dict(self, resource):
        '''
        Return a dict of the give resource, which contains name and resource group.

        :param resource: It can be a resource name, id or a dict contains name and resource group.
        '''
        resource_dict = parse_resource_id(resource) if not isinstance(resource, dict) else resource
        resource_dict['resource_group'] = resource_dict.get('resource_group', self.resource_group)
        if 'subscription_id' in resource_dict:
            resource_dict['subscription_id'] = resource_dict['subscription_id']
        else:
            resource_dict['subscription_id'] = resource_dict.get('subscription', self.subscription_id)
        return resource_dict

    def serialize_obj(self, obj, class_name, enum_modules=None):
        '''
        Return a JSON representation of an Azure object.

        :param obj: Azure object
        :param class_name: Name of the object's class
        :param enum_modules: List of module names to build enum dependencies from.
        :return: serialized result
        '''
        return obj.as_dict()

    def get_poller_result(self, poller, wait=5):
        '''
        Consistent method of waiting on and retrieving results from Azure's long poller

        :param poller Azure poller object
        :return object resulting from the original request
        '''
        try:
            delay = wait
            while not poller.done():
                self.log("Waiting for {0} sec".format(delay))
                poller.wait(timeout=delay)
            return poller.result()
        except Exception as exc:
            self.log(str(exc))
            raise

    def get_multiple_pollers_results(self, pollers, wait=0.05):
        '''
        Consistent method of waiting on and retrieving results from multiple Azure's long poller

        :param pollers list of Azure poller object
        :param wait Period of time to wait for the long running operation to complete.
        :return list of object resulting from the original request
        '''

        def _continue_polling():
            return not all(poller.done() for poller in pollers)

        try:
            while _continue_polling():
                for poller in pollers:
                    if poller.done():
                        continue
                    self.log("Waiting for {0} sec".format(wait))
                    poller.wait(timeout=wait)
            return [poller.result() for poller in pollers]
        except Exception as exc:
            self.log(str(exc))
            raise

    def check_provisioning_state(self, azure_object, requested_state='present'):
        '''
        Check an Azure object's provisioning state. If something did not complete the provisioning
        process, then we cannot operate on it.

        :param azure_object An object such as a subnet, storageaccount, etc. Must have provisioning_state
                            and name attributes.
        :return None
        '''

        if hasattr(azure_object, 'properties') and hasattr(azure_object.properties, 'provisioning_state') and \
           hasattr(azure_object, 'name'):
            # resource group object fits this model
            if isinstance(azure_object.properties.provisioning_state, Enum):
                if azure_object.properties.provisioning_state.value != AZURE_SUCCESS_STATE and \
                   requested_state != 'absent':
                    self.fail("Error {0} has a provisioning state of {1}. Expecting state to be {2}.".format(
                              azure_object.name, azure_object.properties.provisioning_state, AZURE_SUCCESS_STATE))
                return
            if azure_object.properties.provisioning_state != AZURE_SUCCESS_STATE and \
               requested_state != 'absent':
                self.fail("Error {0} has a provisioning state of {1}. Expecting state to be {2}.".format(
                    azure_object.name, azure_object.properties.provisioning_state, AZURE_SUCCESS_STATE))
            return

        if hasattr(azure_object, 'provisioning_state') or not hasattr(azure_object, 'name'):
            if isinstance(azure_object.provisioning_state, Enum):
                if azure_object.provisioning_state.value != AZURE_SUCCESS_STATE and requested_state != 'absent':
                    self.fail("Error {0} has a provisioning state of {1}. Expecting state to be {2}.".format(
                        azure_object.name, azure_object.provisioning_state, AZURE_SUCCESS_STATE))
                return
            if azure_object.provisioning_state != AZURE_SUCCESS_STATE and requested_state != 'absent':
                self.fail("Error {0} has a provisioning state of {1}. Expecting state to be {2}.".format(
                    azure_object.name, azure_object.provisioning_state, AZURE_SUCCESS_STATE))

    def get_blob_service_client(self, resource_group_name, storage_account_name, auth_mode='key'):
        try:
            self.log("Getting storage account detail")
            account = self.storage_client.storage_accounts.get_properties(resource_group_name=resource_group_name, account_name=storage_account_name)
            if auth_mode == 'login' and self.azure_auth.credentials.get('credentials'):
                credential = self.azure_auth.credentials['credentials']
            elif (auth_mode == 'login' and self.azure_auth.credentials.get('tenant')
                  and self.azure_auth.credentials.get('client_id')
                  and self.azure_auth.credentials.get('secret')):
                credential = client_secret.ClientSecretCredential(tenant_id=self.azure_auth.credentials.get('tenant'),
                                                                  client_id=self.azure_auth.credentials.get('client_id'),
                                                                  client_secret=self.azure_auth.credentials.get('secret'))
            else:
                account_keys = self.storage_client.storage_accounts.list_keys(resource_group_name=resource_group_name, account_name=storage_account_name)
                credential = account_keys.keys_property[0].value
        except Exception as exc:
            self.fail("Error getting storage account detail for {0}: {1}".format(storage_account_name, str(exc)))

        try:
            self.log("Create blob service client")
            return BlobServiceClient(
                account_url=account.primary_endpoints.blob,
                credential=credential,
            )
        except Exception as exc:
            self.fail("Error creating blob service client for storage account {0} - {1}".format(storage_account_name, str(exc)))

    def _get_file_share_endpoint_credential(self, resource_group_name, storage_account_name):
        '''
        Resolve the file share data-plane endpoint and account-key credential for the
        given storage account. Used by the file-share directory/file helpers.
        '''
        try:
            account = self.storage_client.storage_accounts.get_properties(resource_group_name=resource_group_name,
                                                                          account_name=storage_account_name)
            account_keys = self.storage_client.storage_accounts.list_keys(resource_group_name=resource_group_name,
                                                                          account_name=storage_account_name)
        except Exception as exc:
            self.fail("Error getting storage account detail for {0}: {1}".format(storage_account_name, str(exc)))
        return account.primary_endpoints.file, account_keys.keys_property[0].value

    def get_share_directory_client(self, resource_group_name, storage_account_name, share_name, directory_path):
        '''
        Build a ShareDirectoryClient for the given directory path inside a file share,
        using the storage account key resolved from the management plane.
        '''
        account_url, credential = self._get_file_share_endpoint_credential(resource_group_name, storage_account_name)
        try:
            return ShareDirectoryClient(account_url=account_url,
                                        share_name=share_name,
                                        directory_path=directory_path,
                                        credential=credential)
        except Exception as exc:
            self.fail("Error creating share directory client for {0}/{1}/{2}: {3}".format(
                storage_account_name, share_name, directory_path, str(exc)))

    def get_share_client(self, resource_group_name, storage_account_name, share_name):
        '''
        Build a ShareClient for the given file share, using the storage account key
        resolved from the management plane.
        '''
        account_url, credential = self._get_file_share_endpoint_credential(resource_group_name, storage_account_name)
        try:
            return ShareClient(account_url=account_url,
                               share_name=share_name,
                               credential=credential)
        except Exception as exc:
            self.fail("Error creating share client for {0}/{1}: {2}".format(
                storage_account_name, share_name, str(exc)))

    def get_share_file_client(self, resource_group_name, storage_account_name, share_name, file_path):
        '''
        Build a ShareFileClient for the given file path inside a file share, using
        the storage account key resolved from the management plane.
        '''
        account_url, credential = self._get_file_share_endpoint_credential(resource_group_name, storage_account_name)
        try:
            return ShareFileClient(account_url=account_url,
                                   share_name=share_name,
                                   file_path=file_path,
                                   credential=credential)
        except Exception as exc:
            self.fail("Error creating share file client for {0}/{1}/{2}: {3}".format(
                storage_account_name, share_name, file_path, str(exc)))

    def create_default_pip(self, resource_group, location, public_ip_name, allocation_method='Dynamic', sku=None):
        '''
        Create a default public IP address <public_ip_name> to associate with a network interface.
        If a PIP address matching <public_ip_name> exists, return it. Otherwise, create one.

        :param resource_group: name of an existing resource group
        :param location: a valid azure location
        :param public_ip_name: base name to assign the public IP address
        :param allocation_method: one of 'Static' or 'Dynamic'
        :param sku: sku
        :return: PIP object
        '''
        pip = None

        self.log("Starting create_default_pip {0}".format(public_ip_name))
        self.log("Check to see if public IP {0} exists".format(public_ip_name))
        try:
            pip = self.network_client.public_ip_addresses.get(resource_group, public_ip_name)
        except Exception:
            pass

        if pip:
            self.log("Public ip {0} found.".format(public_ip_name))
            self.check_provisioning_state(pip)
            return pip

        params = self.network_models.PublicIPAddress(
            location=location,
            public_ip_allocation_method=allocation_method,
            sku=sku
        )
        self.log('Creating default public IP {0}'.format(public_ip_name))
        try:
            poller = self.network_client.public_ip_addresses.begin_create_or_update(resource_group, public_ip_name, params)
        except Exception as exc:
            self.fail("Error creating {0} - {1}".format(public_ip_name, str(exc)))

        return self.get_poller_result(poller)

    def create_default_securitygroup(self, resource_group, location, security_group_name, os_type, open_ports):
        '''
        Create a default security group <security_group_name> to associate with a network interface. If a security group matching
        <security_group_name> exists, return it. Otherwise, create one.

        :param resource_group: Resource group name
        :param location: azure location name
        :param security_group_name: base name to use for the security group
        :param os_type: one of 'Windows' or 'Linux'. Determins any default rules added to the security group.
        :param ssh_port: for os_type 'Linux' port used in rule allowing SSH access.
        :param rdp_port: for os_type 'Windows' port used in rule allowing RDP access.
        :return: security_group object
        '''
        group = None

        self.log("Create security group {0}".format(security_group_name))
        self.log("Check to see if security group {0} exists".format(security_group_name))
        try:
            group = self.network_client.network_security_groups.get(resource_group, security_group_name)
        except Exception:
            pass

        if group:
            self.log("Security group {0} found.".format(security_group_name))
            self.check_provisioning_state(group)
            return group

        parameters = self.network_models.NetworkSecurityGroup()
        parameters.location = location

        if not open_ports:
            # Open default ports based on OS type
            if os_type == 'Linux':
                # add an inbound SSH rule
                parameters.security_rules = [
                    self.network_models.SecurityRule(protocol='Tcp',
                                                     source_address_prefix='*',
                                                     destination_address_prefix='*',
                                                     access='Allow',
                                                     direction='Inbound',
                                                     description='Allow SSH Access',
                                                     source_port_range='*',
                                                     destination_port_range='22',
                                                     priority=100,
                                                     name='SSH')
                ]
                parameters.location = location
            else:
                # for windows add inbound RDP and WinRM rules
                parameters.security_rules = [
                    self.network_models.SecurityRule(protocol='Tcp',
                                                     source_address_prefix='*',
                                                     destination_address_prefix='*',
                                                     access='Allow',
                                                     direction='Inbound',
                                                     description='Allow RDP port 3389',
                                                     source_port_range='*',
                                                     destination_port_range='3389',
                                                     priority=100,
                                                     name='RDP01'),
                    self.network_models.SecurityRule(protocol='Tcp',
                                                     source_address_prefix='*',
                                                     destination_address_prefix='*',
                                                     access='Allow',
                                                     direction='Inbound',
                                                     description='Allow WinRM HTTPS port 5986',
                                                     source_port_range='*',
                                                     destination_port_range='5986',
                                                     priority=101,
                                                     name='WinRM01'),
                ]
        else:
            # Open custom ports
            parameters.security_rules = []
            priority = 100
            for port in open_ports:
                priority += 1
                rule_name = "Rule_{0}".format(priority)
                parameters.security_rules.append(
                    self.network_models.SecurityRule(protocol='Tcp',
                                                     source_address_prefix='*',
                                                     destination_address_prefix='*',
                                                     access='Allow',
                                                     direction='Inbound',
                                                     source_port_range='*',
                                                     destination_port_range=str(port),
                                                     priority=priority,
                                                     name=rule_name)
                )

        self.log('Creating default security group {0}'.format(security_group_name))
        try:
            poller = self.network_client.network_security_groups.begin_create_or_update(resource_group,
                                                                                        security_group_name,
                                                                                        parameters)
        except Exception as exc:
            self.fail("Error creating default security rule {0} - {1}".format(security_group_name, str(exc)))

        return self.get_poller_result(poller)

    @staticmethod
    def _validation_ignore_callback(session, global_config, local_config, **kwargs):
        session.verify = False

    def get_api_profile(self, client_type_name, api_profile_name):
        profile_all_clients = AZURE_API_PROFILES.get(api_profile_name)

        if not profile_all_clients:
            raise KeyError("unknown Azure API profile: {0}".format(api_profile_name))

        profile_raw = profile_all_clients.get(client_type_name, None)

        if not profile_raw:
            self.module.warn("Azure API profile {0} does not define an entry for {1}".format(api_profile_name, client_type_name))

        if isinstance(profile_raw, dict):
            if not profile_raw.get('default_api_version'):
                raise KeyError("Azure API profile {0} does not define 'default_api_version'".format(api_profile_name))
            return profile_raw

        # wrap basic strings in a dict that just defines the default
        return dict(default_api_version=profile_raw)

    # The graphrbac has deprecated, migrate to msgraph
    # def get_graphrbac_client(self, tenant_id):
    #    cred = self.azure_auth.azure_credentials
    #    base_url = self.azure_auth._cloud_environment.endpoints.active_directory_graph_resource_id
    #    client = GraphRbacManagementClient(cred, tenant_id, base_url)
    #    return client

    def get_msgraph_client(self):
        auth_provider = AzureIdentityAuthenticationProvider(self.azure_auth.azure_credential_track2)
        cloud_mapping = {
            azure_cloud.AZURE_CHINA_CLOUD: NationalClouds.China,
            azure_cloud.AZURE_US_GOV_CLOUD: NationalClouds.US_GOV,
            azure_cloud.AZURE_GERMAN_CLOUD: NationalClouds.Germany
        }
        host = cloud_mapping.get(self._cloud_environment, NationalClouds.Global)
        client = GraphClientFactory.create_with_default_middleware(host=host)
        request_adapter = GraphRequestAdapter(auth_provider, client=client)
        return GraphServiceClient(self.azure_auth.azure_credential_track2, request_adapter=request_adapter)

    def get_mgmt_svc_client(self, client_type, base_url=None, api_version=None, suppress_subscription_id=False):
        self.log('Getting management service client {0}'.format(client_type.__name__))

        client_argspec = inspect.signature(client_type.__init__)

        if not base_url:
            # most things are resource_manager, don't make everyone specify
            base_url = self.azure_auth._cloud_environment.endpoints.resource_manager

        # https://github.com/Azure/msrestazure-for-python/pull/169
        # China's base_url doesn't end in a trailing slash, though others do,
        # and we need a trailing slash when generating credential_scopes below.
        if not base_url.endswith("/"):
            base_url += "/"

        mgmt_subscription_id = self.subscription_id

        # Some management clients do not take a subscription ID as parameters.
        if suppress_subscription_id:
            client_kwargs = dict(credential=self.azure_auth.azure_credential_track2, base_url=base_url, credential_scopes=[base_url + ".default"])
        else:
            client_kwargs = dict(credential=self.azure_auth.azure_credential_track2,
                                 subscription_id=mgmt_subscription_id, base_url=base_url, credential_scopes=[base_url + ".default"])

        api_profile_dict = {}

        if self.api_profile:
            api_profile_dict = self.get_api_profile(client_type.__name__, self.api_profile)

        # unversioned clients won't accept profile; only send it if necessary
        # clients without a version specified in the profile will use the default
        if api_profile_dict and 'profile' in client_argspec.parameters:
            client_kwargs['profile'] = api_profile_dict

        # If the client doesn't accept api_version, it's unversioned.
        # If it does, favor explicitly-specified api_version, fall back to api_profile
        if 'api_version' in client_argspec.parameters:
            profile_default_version = api_profile_dict.get('default_api_version', None)
            if api_version or profile_default_version:
                client_kwargs['api_version'] = api_version or profile_default_version
                if 'profile' in client_kwargs:
                    # remove profile; only pass API version if specified
                    client_kwargs.pop('profile')

        client = client_type(**client_kwargs)

        # FUTURE: remove this once everything exposes models directly (eg, containerinstance)
        try:
            getattr(client, "models")
        except AttributeError:
            def _ansible_get_models(self, *arg, **kwarg):
                return self._ansible_models

            try:
                setattr(client, '_ansible_models', importlib.import_module(client_type.__module__).models)
            except AttributeError:
                if hasattr(importlib.import_module(client_type.__module__), '_modeles'):
                    setattr(client, '_ansible_models', importlib.import_module(client_type.__module__)._models)
                else:
                    pass
            client.models = types.MethodType(_ansible_get_models, client)

        client._config.user_agent_policy = self.add_user_agent(client._config.user_agent_policy)

        if self.azure_auth._cert_validation_mode == 'ignore':
            client._config.session_configuration_callback = self._validation_ignore_callback

        return client

    def add_user_agent(self, config):
        # Add user agent for Ansible
        config.add_user_agent(ANSIBLE_USER_AGENT)
        # Add user agent when running from Cloud Shell
        if CLOUDSHELL_USER_AGENT_KEY in os.environ:
            config.add_user_agent(os.environ[CLOUDSHELL_USER_AGENT_KEY])
        # Add user agent when running from VSCode extension
        if VSCODEEXT_USER_AGENT_KEY in os.environ:
            config.add_user_agent(os.environ[VSCODEEXT_USER_AGENT_KEY])
        return config

    def generate_sas_token(self, **kwags):
        base_url = kwags.get('base_url', None)
        expiry = kwags.get('expiry', time() + 3600)
        key = kwags.get('key', None)
        policy = kwags.get('policy', None)
        url = quote_plus(base_url)
        ttl = int(expiry)
        sign_key = '{0}\n{1}'.format(url, ttl)
        signature = b64encode(HMAC(b64decode(key), sign_key.encode('utf-8'), sha256).digest())
        result = {
            'sr': url,
            'sig': signature,
            'se': str(ttl),
        }
        if policy:
            result['skn'] = policy
        return 'SharedAccessSignature ' + urlencode(result)

    def get_subnet_detail(self, subnet_id):
        vnet_detail = subnet_id.split('/Microsoft.Network/virtualNetworks/')[1].split('/subnets/')
        return dict(
            resource_group=subnet_id.split('resourceGroups/')[1].split('/')[0],
            vnet_name=vnet_detail[0],
            subnet_name=vnet_detail[1],
        )

    # passthru methods to AzureAuth instance for backcompat
    @property
    def credentials(self):
        return self.azure_auth.credentials

    @property
    def _cloud_environment(self):
        return self.azure_auth._cloud_environment

    @property
    def subscription_id(self):
        return self.azure_auth.subscription_id

    @property
    def storage_client(self):
        self.log('Getting storage client...')
        if not self._storage_client:
            self._storage_client = self.get_mgmt_svc_client(StorageManagementClient,
                                                            base_url=self._cloud_environment.endpoints.resource_manager,
                                                            api_version='2025-08-01')
        return self._storage_client

    @property
    def storage_models(self):
        return StorageModels

    @property
    def authorization_client(self):
        self.log('Getting authorization client...')
        if not self._authorization_client:
            self._authorization_client = self.get_mgmt_svc_client(AuthorizationManagementClient,
                                                                  base_url=self._cloud_environment.endpoints.resource_manager,
                                                                  api_version='2020-04-01-preview')
        return self._authorization_client

    @property
    def authorization_models(self):
        return AuthorizationManagementClient.models('2020-04-01-preview')

    @property
    def subscription_client(self):
        self.log('Getting subscription client...')
        if not self._subscription_client:
            self._subscription_client = self.get_mgmt_svc_client(SubscriptionClient,
                                                                 base_url=self._cloud_environment.endpoints.resource_manager,
                                                                 suppress_subscription_id=True,
                                                                 api_version='2022-12-01')
        return self._subscription_client

    @property
    def subscription_models(self):
        return SubscriptionClient.models("2022-12-01")

    @property
    def management_groups_client(self):
        self.log('Getting Management Groups client...')
        if not self._management_group_client:
            self._management_group_client = self.get_mgmt_svc_client(ManagementGroupsClient,
                                                                     base_url=self._cloud_environment.endpoints.resource_manager,
                                                                     suppress_subscription_id=True,
                                                                     api_version='2020-05-01')
        return self._management_group_client

    @property
    def network_client(self):
        self.log('Getting network client')
        if not self._network_client:
            self._network_client = self.get_mgmt_svc_client(NetworkManagementClient,
                                                            base_url=self._cloud_environment.endpoints.resource_manager,
                                                            api_version='2024-01-01')
        return self._network_client

    @property
    def network_models(self):
        self.log("Getting network models...")
        return NetworkModels

    @property
    def rm_client(self):
        self.log('Getting resource manager client')
        if not self._resource_client:
            self._resource_client = self.get_mgmt_svc_client(ResourceManagementClient,
                                                             base_url=self._cloud_environment.endpoints.resource_manager,
                                                             api_version='2022-09-01')
        return self._resource_client

    @property
    def rm_models(self):
        self.log("Getting resource manager models")
        return ResourceManagementClient.models("2022-09-01")

    @property
    def image_client(self):
        self.log('Getting compute image client')
        if not self._image_client:
            self._image_client = self.get_mgmt_svc_client(ComputeManagementClient,
                                                          base_url=self._cloud_environment.endpoints.resource_manager,
                                                          api_version='2024-07-01')
        return self._image_client

    @property
    def image_models(self):
        self.log("Getting compute image models")
        return ComputeManagementClient.models("2024-07-01")

    @property
    def compute_client(self):
        self.log('Getting compute client')
        if not self._compute_client:
            self._compute_client = self.get_mgmt_svc_client(ComputeManagementClient,
                                                            base_url=self._cloud_environment.endpoints.resource_manager,
                                                            api_version='2024-07-01')
        return self._compute_client

    @property
    def compute_models(self):
        self.log("Getting compute models")
        return ComputeManagementClient.models("2024-07-01")

    @property
    def disk_client(self):
        self.log('Getting disk client')
        if not self._disk_client:
            self._disk_client = self.get_mgmt_svc_client(ComputeManagementClient,
                                                         base_url=self._cloud_environment.endpoints.resource_manager,
                                                         api_version='2023-04-02')
        return self._disk_client

    @property
    def disk_models(self):
        self.log("Getting disk models")
        return ComputeManagementClient.models("2023-04-02")

    @property
    def image_version_client(self):
        self.log('Getting gallery image version client')
        if not self._image_version_client:
            self._image_version_client = self.get_mgmt_svc_client(ComputeManagementClient,
                                                                  base_url=self._cloud_environment.endpoints.resource_manager,
                                                                  api_version='2023-07-03')
        return self._image_version_client

    @property
    def image_version_models(self):
        self.log("Getting image version models")
        return ComputeManagementClient.models("2023-07-03")

    @property
    def multi_disk_client(self):
        self.log('Getting disk client')
        if not self._multi_disk_client:
            self._multi_disk_client = self.get_mgmt_svc_client(ComputeManagementClient,
                                                               base_url=self._cloud_environment.endpoints.resource_manager,
                                                               api_version='2021-04-01')
        return self._multi_disk_client

    @property
    def multi_disk_models(self):
        self.log("Getting disk models")
        return ComputeManagementClient.models("2021-04-01")

    @property
    def diskencryptionset_client(self):
        self.log('Getting diskencryptionset client')
        base_url = self._cloud_environment.endpoints.resource_manager
        if not self._diskencryptionset_client:
            self._diskencryptionset_client = self.get_mgmt_svc_client(ComputeManagementClient,
                                                                      base_url=base_url,
                                                                      api_version='2024-03-02')
        return self._diskencryptionset_client

    @property
    def diskencryptionset_models(self):
        self.log("Getting compute models")
        return ComputeManagementClient.models("2024-03-02")

    @property
    def dns_client(self):
        self.log('Getting dns client')
        if not self._dns_client:
            self._dns_client = self.get_mgmt_svc_client(DnsManagementClient,
                                                        base_url=self._cloud_environment.endpoints.resource_manager,
                                                        api_version='2023-07-01-preview')
        return self._dns_client

    @property
    def dns_models(self):
        self.log("Getting dns models...")
        return DnsManagementClient.models('2023-07-01-preview')

    @property
    def private_dns_client(self):
        self.log('Getting private dns client')
        if not self._private_dns_client:
            self._private_dns_client = self.get_mgmt_svc_client(
                PrivateDnsManagementClient,
                base_url=self._cloud_environment.endpoints.resource_manager)
        return self._private_dns_client

    @property
    def private_dns_models(self):
        self.log('Getting private dns models')
        return PrivateDnsModels

    @property
    def web_client(self):
        self.log('Getting web client')
        if not self._web_client:
            self._web_client = self.get_mgmt_svc_client(WebSiteManagementClient,
                                                        base_url=self._cloud_environment.endpoints.resource_manager,
                                                        api_version='2023-12-01')
        return self._web_client

    @property
    def containerservice_client(self):
        self.log('Getting container service client')
        if not self._containerservice_client:
            self._containerservice_client = self.get_mgmt_svc_client(ContainerServiceClient,
                                                                     base_url=self._cloud_environment.endpoints.resource_manager,
                                                                     api_version='2019-04-01')
        return self._containerservice_client

    @property
    def managedcluster_models(self):
        self.log("Getting container service models")
        return ContainerServiceClient.models('2024-08-01')

    @property
    def managedcluster_client(self):
        self.log('Getting container service client')
        if not self._managedcluster_client:
            self._managedcluster_client = self.get_mgmt_svc_client(ContainerServiceClient,
                                                                   base_url=self._cloud_environment.endpoints.resource_manager,
                                                                   api_version='2024-08-01')
        return self._managedcluster_client

    @property
    def sql_client(self):
        self.log('Getting SQL client')
        if not self._sql_client:
            self._sql_client = self.get_mgmt_svc_client(SqlManagementClient,
                                                        base_url=self._cloud_environment.endpoints.resource_manager)
        return self._sql_client

    @property
    def postgresql_flexible_client(self):
        self.log('Getting PostgreSQL client')
        if not self._postgresql_flexible_client:
            self._postgresql_flexible_client = self.get_mgmt_svc_client(PostgreSQLFlexibleManagementClient,
                                                                        base_url=self._cloud_environment.endpoints.resource_manager,
                                                                        api_version='2025-08-01')
        return self._postgresql_flexible_client

    @property
    def postgresql_client(self):
        self.log('Getting PostgreSQL client')
        if not self._postgresql_client:
            self._postgresql_client = self.get_mgmt_svc_client(PostgreSQLManagementClient,
                                                               base_url=self._cloud_environment.endpoints.resource_manager)
        return self._postgresql_client

    @property
    def mysql_flexible_client(self):
        self.log('Getting MySQL Flexible client')
        if not self._mysql_flexible_client:
            self._mysql_flexible_client = self.get_mgmt_svc_client(MySQLFlexibleManagementClient,
                                                                   base_url=self._cloud_environment.endpoints.resource_manager,
                                                                   api_version='2024-08-01')
        return self._mysql_flexible_client

    @property
    def mysql_client(self):
        self.log('Getting MySQL client')
        if not self._mysql_client:
            self._mysql_client = self.get_mgmt_svc_client(MySQLManagementClient,
                                                          base_url=self._cloud_environment.endpoints.resource_manager)
        return self._mysql_client

    @property
    def mariadb_client(self):
        self.log('Getting MariaDB client')
        if not self._mariadb_client:
            self._mariadb_client = self.get_mgmt_svc_client(MariaDBManagementClient,
                                                            base_url=self._cloud_environment.endpoints.resource_manager)
        return self._mariadb_client

    @property
    def containerregistry_client(self):
        self.log('Getting container registry mgmt client')
        if not self._containerregistry_client:
            self._containerregistry_client = self.get_mgmt_svc_client(ContainerRegistryManagementClient,
                                                                      base_url=self._cloud_environment.endpoints.resource_manager,
                                                                      api_version='2021-09-01')

        return self._containerregistry_client

    @property
    def containerregistrytoken_client(self):
        self.log('Getting container registry token mgmt client')
        if not self._containerregistrytoken_client:
            self._containerregistrytoken_client = self.get_mgmt_svc_client(ContainerRegistryManagementClient,
                                                                           base_url=self._cloud_environment.endpoints.resource_manager,
                                                                           api_version='2023-07-01')

        return self._containerregistrytoken_client

    @property
    def containerinstance_client(self):
        self.log('Getting container instance mgmt client')
        if not self._containerinstance_client:
            self._containerinstance_client = self.get_mgmt_svc_client(ContainerInstanceManagementClient,
                                                                      base_url=self._cloud_environment.endpoints.resource_manager,
                                                                      api_version='2023-05-01')

        return self._containerinstance_client

    @property
    def marketplace_client(self):
        self.log('Getting marketplace agreement client')
        if not self._marketplace_client:
            self._marketplace_client = self.get_mgmt_svc_client(MarketplaceOrderingAgreements,
                                                                base_url=self._cloud_environment.endpoints.resource_manager)
        return self._marketplace_client

    @property
    def traffic_manager_management_client(self):
        self.log('Getting traffic manager client')
        if not self._traffic_manager_management_client:
            self._traffic_manager_management_client = self.get_mgmt_svc_client(TrafficManagerManagementClient,
                                                                               base_url=self._cloud_environment.endpoints.resource_manager)
        return self._traffic_manager_management_client

    @property
    def monitor_autoscale_settings_client(self):
        self.log('Getting monitor client for autoscale_settings')
        if not self._monitor_autoscale_settings_client:
            self._monitor_autoscale_settings_client = self.get_mgmt_svc_client(MonitorManagementClient,
                                                                               base_url=self._cloud_environment.endpoints.resource_manager,
                                                                               api_version="2022-10-01")
        return self._monitor_autoscale_settings_client

    @property
    def monitor_log_profiles_client(self):
        self.log('Getting monitor client for log_profiles')
        if not self._monitor_log_profiles_client:
            self._monitor_log_profiles_client = self.get_mgmt_svc_client(MonitorManagementClient,
                                                                         base_url=self._cloud_environment.endpoints.resource_manager,
                                                                         api_version="2016-03-01")
        return self._monitor_log_profiles_client

    @property
    def monitor_diagnostic_settings_client(self):
        self.log('Getting monitor client for diagnostic_settings')
        if not self._monitor_diagnostic_settings_client:
            self._monitor_diagnostic_settings_client = self.get_mgmt_svc_client(MonitorManagementClient,
                                                                                base_url=self._cloud_environment.endpoints.resource_manager,
                                                                                api_version="2021-05-01-preview")
        return self._monitor_diagnostic_settings_client

    @property
    def monitor_management_client_action_groups(self):
        self.log('Getting monitor client for diagnostic_settings')
        if not self._monitor_management_client_action_groups:
            self._monitor_management_client_action_groups = self.get_mgmt_svc_client(MonitorManagementClient,
                                                                                     base_url=self._cloud_environment.endpoints.resource_manager,
                                                                                     api_version='2023-01-01')
        return self._monitor_management_client_action_groups

    @property
    def monitor_management_client_activity_log_alerts(self):
        self.log('Getting monitor client for diagnostic_settings')
        if not self._monitor_management_client_activity_log_alerts:
            self._monitor_management_client_activity_log_alerts = self.get_mgmt_svc_client(MonitorManagementClient,
                                                                                           base_url=self._cloud_environment.endpoints.resource_manager,
                                                                                           api_version='2020-10-01')
        return self._monitor_management_client_activity_log_alerts

    @property
    def monitor_management_client_data_collection_rules(self):
        self.log('Getting monitor client for diagnostic_settings')
        if not self._monitor_data_collection_rules_client:
            self._monitor_data_collection_rules_client = self.get_mgmt_svc_client(MonitorManagementClient,
                                                                                  base_url=self._cloud_environment.endpoints.resource_manager,
                                                                                  api_version='2022-06-01')
        return self._monitor_data_collection_rules_client

    @property
    def monitor_management_client_metric_alerts(self):
        self.log('Getting monitor client for diagnostic_settings')
        if not self._monitor_management_client_metric_alerts:
            self._monitor_management_client_metric_alerts = self.get_mgmt_svc_client(MonitorManagementClient,
                                                                                     base_url=self._cloud_environment.endpoints.resource_manager,
                                                                                     api_version='2018-03-01')
        return self._monitor_management_client_metric_alerts

    @property
    def monitor_management_client_scheduled_query_rules(self):
        self.log('Getting monitor client for scheduled query rules')
        if not self._monitor_management_client_scheduled_query_rules:
            self._monitor_management_client_scheduled_query_rules = self.get_mgmt_svc_client(MonitorManagementClient,
                                                                                             base_url=self._cloud_environment.endpoints.resource_manager,
                                                                                             api_version='2018-04-16')
        return self._monitor_management_client_scheduled_query_rules

    @property
    def log_analytics_client(self):
        self.log('Getting log analytics client')
        if not self._log_analytics_client:
            self._log_analytics_client = self.get_mgmt_svc_client(LogAnalyticsManagementClient,
                                                                  base_url=self._cloud_environment.endpoints.resource_manager)
        return self._log_analytics_client

    @property
    def log_analytics_models(self):
        self.log('Getting log analytics models')
        return LogAnalyticsModels

    @property
    def servicebus_client(self):
        self.log('Getting servicebus client')
        if not self._servicebus_client:
            self._servicebus_client = self.get_mgmt_svc_client(ServiceBusManagementClient,
                                                               api_version="2022-10-01-preview",
                                                               base_url=self._cloud_environment.endpoints.resource_manager)
        return self._servicebus_client

    @property
    def servicebus_models(self):
        return ServiceBusManagementClient.models("2022-10-01-preview")

    @property
    def automation_client(self):
        self.log('Getting automation client')
        if not self._automation_client:
            self._automation_client = self.get_mgmt_svc_client(AutomationClient,
                                                               base_url=self._cloud_environment.endpoints.resource_manager)
        return self._automation_client

    @property
    def automation_models(self):
        return AutomationModel

    @property
    def IoThub_client(self):
        self.log('Getting iothub client')
        if not self._IoThub_client:
            self._IoThub_client = self.get_mgmt_svc_client(IotHubClient,
                                                           api_version='2023-06-30-preview',
                                                           base_url=self._cloud_environment.endpoints.resource_manager)
        return self._IoThub_client

    @property
    def IoThub_models(self):
        return IoTHubModels

    @property
    def lock_client(self):
        self.log('Getting lock client')
        if not self._lock_client:
            self._lock_client = self.get_mgmt_svc_client(ManagementLockClient,
                                                         base_url=self._cloud_environment.endpoints.resource_manager,
                                                         api_version='2016-09-01')
        return self._lock_client

    @property
    def lock_models(self):
        self.log("Getting lock models")
        return ManagementLockClient.models('2016-09-01')

    @property
    def recovery_services_backup_client(self):
        self.log('Getting recovery services backup client')
        if not self._recovery_services_backup_client:
            self._recovery_services_backup_client = self.get_mgmt_svc_client(RecoveryServicesBackupClient,
                                                                             base_url=self._cloud_environment.endpoints.resource_manager)
        return self._recovery_services_backup_client

    @property
    def recovery_services_backup_models(self):
        return RecoveryServicesBackupModels

    @property
    def search_client(self):
        self.log('Getting search client...')
        if not self._search_client:
            self._search_client = self.get_mgmt_svc_client(SearchManagementClient,
                                                           base_url=self._cloud_environment.endpoints.resource_manager,
                                                           api_version='2024-06-01-preview')
        return self._search_client

    @property
    def notification_hub_client(self):
        self.log('Getting notification hub client')
        if not self._notification_hub_client:
            self._notification_hub_client = self.get_mgmt_svc_client(
                NotificationHubsManagementClient,
                base_url=self._cloud_environment.endpoints.resource_manager,
                api_version='2023-09-01')
        return self._notification_hub_client

    @property
    def event_hub_client(self):
        self.log('Getting event hub client')
        if not self._event_hub_client:
            self._event_hub_client = self.get_mgmt_svc_client(
                EventHubManagementClient,
                base_url=self._cloud_environment.endpoints.resource_manager,
                api_version='2024-01-01')
        return self._event_hub_client

    @property
    def datafactory_client(self):
        self.log('Getting datafactory client...')
        if not self._datafactory_client:
            self._datafactory_client = self.get_mgmt_svc_client(DataFactoryManagementClient,
                                                                base_url=self._cloud_environment.endpoints.resource_manager)
        return self._datafactory_client

    @property
    def datafactory_model(self):
        return DataFactoryModel

    @property
    def batch_account_client(self):
        self.log('Getting batch account client...')
        if not self._batch_account_client:
            self._batch_account_client = self.get_mgmt_svc_client(BatchManagementClient,
                                                                  base_url=self._cloud_environment.endpoints.resource_manager)
        return self._batch_account_client

    @property
    def resourcehealth_client(self):
        self.log('Getting resource health client...')
        if not self._resourcehealth_client:
            self._resourcehealth_client = self.get_mgmt_svc_client(ResourceHealthMgmtClient,
                                                                   base_url=self._cloud_environment.endpoints.resource_manager)
        return self._resourcehealth_client

    @property
    def batch_account_model(self):
        return BatchManagementModel

    @property
    def cdn_client(self):
        self.log('Getting cdn client...')
        if not self._cdn_client:
            self._cdn_client = self.get_mgmt_svc_client(CdnManagementClient,
                                                        base_url=self._cloud_environment.endpoints.resource_manager,
                                                        api_version='2024-02-01')
        return self._cdn_client

    @property
    def hybrid_compute_management_client(self):
        self.log('Getting hybrid compute management client...')
        if not self._hybrid_compute_management_client:
            self._hybrid_compute_management_client = self.get_mgmt_svc_client(HybridComputeManagementClient,
                                                                              base_url=self._cloud_environment.endpoints.resource_manager)
        return self._hybrid_compute_management_client


class AzureRMAuthException(Exception):
    pass


class AzureRMAuth(object):
    _cloud_environment = None
    _adfs_authority_url = None

    def __init__(self, auth_source=None, profile=None, subscription_id=None, client_id=None, secret=None,
                 tenant=None, ad_user=None, password=None, cloud_environment='AzureCloud', cert_validation_mode='validate',
                 api_profile='latest', adfs_authority_url=None, fail_impl=None, is_ad_resource=False,
                 x509_certificate_path=None, thumbprint=None, oidc_token=None, oidc_token_file_path=None, track1_cred=False,
                 disable_instance_discovery=False , **kwargs):

        if fail_impl:
            self._fail_impl = fail_impl
        else:
            self._fail_impl = self._default_fail_impl
        self.is_ad_resource = is_ad_resource

        # authenticate
        self.credentials = self._get_credentials(
            auth_source=auth_source,
            profile=profile,
            subscription_id=subscription_id,
            client_id=client_id,
            secret=secret,
            tenant=tenant,
            ad_user=ad_user,
            password=password,
            cloud_environment=cloud_environment,
            cert_validation_mode=cert_validation_mode,
            api_profile=api_profile,
            adfs_authority_url=adfs_authority_url,
            x509_certificate_path=x509_certificate_path,
            thumbprint=thumbprint,
            disable_instance_discovery=disable_instance_discovery,
            oidc_token=oidc_token,
            oidc_token_file_path=oidc_token_file_path)

        if not self.credentials:
            self.fail("Failed to get credentials. Either pass as parameters, set environment variables, "
                      "define a profile in ~/.azure/credentials, or log in with Azure CLI (`az login`).")

        # cert validation mode precedence: module-arg, credential profile, env, "validate"
        self._cert_validation_mode = cert_validation_mode or \
            self.credentials.get('cert_validation_mode') or \
            self._get_env('cert_validation_mode') or \
            'validate'

        if self._cert_validation_mode not in ['validate', 'ignore']:
            self.fail('invalid cert_validation_mode: {0}'.format(self._cert_validation_mode))

        # Disable instance discovery: module-arg, credential profile, env, "False"
        self._disable_instance_discovery = disable_instance_discovery or \
            self.credentials.get('disable_instance_discovery') or \
            self._get_env('disable_instance_discovery') or \
            False

        # if cloud_environment specified, look up/build Cloud object
        raw_cloud_env = self.credentials.get('cloud_environment')
        if self.credentials.get('credentials') is not None and raw_cloud_env is not None:
            self._cloud_environment = raw_cloud_env
        elif not raw_cloud_env:
            self._cloud_environment = azure_cloud.AZURE_PUBLIC_CLOUD  # SDK default
        else:
            # try to look up "well-known" values via the name attribute on azure_cloud members
            all_clouds = [x[1] for x in inspect.getmembers(azure_cloud) if isinstance(x[1], azure_cloud.Cloud)]
            matched_clouds = [x for x in all_clouds if x.name == raw_cloud_env]
            if len(matched_clouds) == 1:
                self._cloud_environment = matched_clouds[0]
            elif len(matched_clouds) > 1:
                self.fail("Azure SDK failure: more than one cloud matched for cloud_environment name '{0}'".format(raw_cloud_env))
            else:
                if not urlparse.urlparse(raw_cloud_env).scheme:
                    self.fail("cloud_environment must be an endpoint discovery URL or one of {0}".format([x.name for x in all_clouds]))
                try:
                    self._cloud_environment = self._get_cloud_from_metadata_endpoint(raw_cloud_env)
                except Exception as e:
                    self.fail("cloud_environment {0} could not be resolved: {1}".format(raw_cloud_env, e.message), exception=traceback.format_exc())

        if self.credentials.get('subscription_id', None) is None and not self.is_ad_resource:
            self.fail("Credentials did not include a subscription_id value.")
        self.log("setting subscription_id")
        self.subscription_id = self.credentials.get('subscription_id')

        # get authentication authority
        # for adfs, user could pass in authority or not.
        # for others, use default authority from cloud environment
        if self.credentials.get('adfs_authority_url') is None:
            self._adfs_authority_url = self._cloud_environment.endpoints.active_directory
        else:
            self._adfs_authority_url = self.credentials.get('adfs_authority_url')

        if self.credentials.get('auth_source') == 'msi':
            # MSI Credentials
            self.azure_credential_track2 = self.credentials['credentials']
        elif self.credentials.get('credentials') is not None:
            # AzureCLI credentials
            self.azure_credential_track2 = self.credentials['credentials']
        # OIDC direct token
        elif self.credentials.get('client_id') is not None and \
                self.credentials.get('tenant') is not None and \
                self.credentials.get('oidc_token') is not None:
            token = self.credentials['oidc_token']

            def _load_assertion():
                return token

            self.azure_credential_track2 = ClientAssertionCredential(tenant_id=self.credentials['tenant'],
                                                                     client_id=self.credentials['client_id'],
                                                                     func=_load_assertion,
                                                                     authority=self._adfs_authority_url,
                                                                     disable_instance_discovery=self._disable_instance_discovery)
        # OIDC token file
        elif self.credentials.get('client_id') is not None and \
                self.credentials.get('tenant') is not None and \
                self.credentials.get('oidc_token_file_path') is not None:
            token_file = self.credentials['oidc_token_file_path']

            if not os.path.exists(token_file):
                self.fail(f"The specified OIDC token file does not exist: {token_file}")

            def _load_assertion():
                with open(token_file) as f:
                    return f.read()

            self.azure_credential_track2 = ClientAssertionCredential(tenant_id=self.credentials['tenant'],
                                                                     client_id=self.credentials['client_id'],
                                                                     func=_load_assertion,
                                                                     authority=self._adfs_authority_url,
                                                                     disable_instance_discovery=self._disable_instance_discovery)

        elif self.credentials.get('client_id') is not None and \
                self.credentials.get('secret') is not None and \
                self.credentials.get('tenant') is not None:
            self.azure_credential_track2 = client_secret.ClientSecretCredential(client_id=self.credentials['client_id'],
                                                                                client_secret=self.credentials['secret'],
                                                                                tenant_id=self.credentials['tenant'],
                                                                                authority=self._adfs_authority_url,
                                                                                disable_instance_discovery=self._disable_instance_discovery)

        elif self.credentials.get('client_id') is not None and \
                self.credentials.get('tenant') is not None and \
                self.credentials.get('thumbprint') is not None and \
                self.credentials.get('x509_certificate_path') is not None:
            self.azure_credential_track2 = certificate.CertificateCredential(tenant_id=self.credentials['tenant'],
                                                                             client_id=self.credentials['client_id'],
                                                                             certificate_path=self.credentials['x509_certificate_path'],
                                                                             authority=self._adfs_authority_url,
                                                                             disable_instance_discovery=self._disable_instance_discovery)

        elif self.credentials.get('ad_user') is not None and \
                self.credentials.get('password') is not None and \
                self.credentials.get('client_id') is not None and \
                self.credentials.get('tenant') is not None:
            self.azure_credential_track2 = user_password.UsernamePasswordCredential(username=self.credentials['ad_user'],
                                                                                    password=self.credentials['password'],
                                                                                    tenant_id=self.credentials.get('tenant'),
                                                                                    client_id=self.credentials.get('client_id'),
                                                                                    authority=self._adfs_authority_url,
                                                                                    disable_instance_discovery=self._disable_instance_discovery)

        elif self.credentials.get('ad_user') is not None and self.credentials.get('password') is not None:
            client_id = self.credentials.get('client_id')
            if client_id is None:
                client_id = '04b07795-8ddb-461a-bbee-02f9e1bf7b46'
            self.azure_credential_track2 = user_password.UsernamePasswordCredential(username=self.credentials['ad_user'],
                                                                                    password=self.credentials['password'],
                                                                                    tenant_id=self.credentials.get('tenant', 'organizations'),
                                                                                    client_id=client_id,
                                                                                    authority=self._adfs_authority_url,
                                                                                    disable_instance_discovery=self._disable_instance_discovery)

        else:
            self.fail("Failed to authenticate with provided credentials. Some attributes were missing. \n\n"
                      "Supported authentication methods: \n"
                      "- Workload identity (OIDC) token: client_id, tenant, and oidc_token\n"
                      "- Workload identity (OIDC) token file: client_id, tenant, and oidc_token_file_path\n"
                      "- Service principal with client secret: client_id, tenant, secret\n"
                      "- Service principal with certificate: client_id, tenant, x509_certificate_path\n"
                      "- Username/password authentication: ad_user, password\n"
                      "- Azure CLI authentication: run `az_login`.\n")

    def fail(self, msg, exception=None, **kwargs):
        self._fail_impl(msg)

    def _default_fail_impl(self, msg, exception=None, **kwargs):
        raise AzureRMAuthException(msg)

    def _get_env(self, module_key, default=None):
        "Read envvar matching module parameter"
        return os.environ.get(AZURE_CREDENTIAL_ENV_MAPPING[module_key], default)

    def _get_profile(self, profile="default", subscription_id=None):
        path = expanduser("~/.azure/credentials")
        try:
            config = configparser.ConfigParser()
            config.read(path)
        except Exception as exc:
            self.fail("Failed to access {0}. Check that the file exists and you have read "
                      "access. {1}".format(path, str(exc)))
        credentials = dict()
        for key in AZURE_CREDENTIAL_ENV_MAPPING:
            try:
                credentials[key] = config.get(profile, key, raw=True)
            except Exception:
                pass

        if credentials.get('subscription_id') is None and not self.is_ad_resource:
            return None

        if subscription_id is not None:
            credentials['subscription_id'] = subscription_id

        return credentials

    def _get_msi_credentials(self, subscription_id=None, client_id=None, _cloud_environment=None, **kwargs):
        # Get object `cloud_environment` from string `_cloud_environment`
        cloud_environment = None
        if not _cloud_environment:
            cloud_environment = azure_cloud.AZURE_PUBLIC_CLOUD
        else:
            # try to look up "well-known" values via the name attribute on azure_cloud members
            all_clouds = [x[1] for x in inspect.getmembers(azure_cloud) if isinstance(x[1], azure_cloud.Cloud)]
            matched_clouds = [x for x in all_clouds if x.name == _cloud_environment]
            if len(matched_clouds) == 1:
                cloud_environment = matched_clouds[0]
            elif len(matched_clouds) > 1:
                self.fail("Azure SDK failure: more than one cloud matched for cloud_environment name '{0}'".format(_cloud_environment))
            else:
                if not urlparse.urlparse(_cloud_environment).scheme:
                    self.fail("cloud_environment must be an endpoint discovery URL or one of {0}".format([x.name for x in all_clouds]))
                try:
                    cloud_environment = self._get_cloud_from_metadata_endpoint(_cloud_environment)
                except Exception as exc:
                    self.fail("cloud_environment {0} could not be resolved: {1}".format(_cloud_environment, str(exc)), exception=traceback.format_exc())

        client_id = client_id or self._get_env('client_id')
        credential = managed_identity.ManagedIdentityCredential(client_id=client_id, cloud_environment=cloud_environment)
        subscription_id = subscription_id or self._get_env('subscription_id')
        if not subscription_id:
            try:
                # use the first subscription of the MSI
                subscription_client = SubscriptionClient(credential)
                subscription = next(subscription_client.subscriptions.list())
                subscription_id = str(subscription.subscription_id)
            except Exception as exc:
                self.fail("Failed to get MSI token: {0}. "
                          "Please check whether your machine enabled MSI or grant access to any subscription.".format(str(exc)))
        return {
            'credentials': credential,
            'subscription_id': subscription_id,
            'cloud_environment': cloud_environment,
            'auth_source': 'msi'
        }

    def _get_azure_cli_credentials(self, subscription_id=None):
        subscription_id = subscription_id or self._get_env('subscription_id')
        if not subscription_id:
            try:
                cmd = ["az", "account", "show", "--query", "id", "-o", "json"]
                subscription_id = subprocess.run(cmd, capture_output=True, text=True, check=True).stdout.strip().strip('"')  # allow subprocess
            except Exception as ec:
                raise CLIError("Obtain the az login's subscription occurred exception as {0}".format(ec))

        try:
            cmd = ["az", "cloud", "show", "--query", "name", "-o", "json"]
            cloud_name = subprocess.run(cmd, capture_output=True, text=True, check=True).stdout.strip().strip('"')  # allow subprocess
            all_clouds = [x[1] for x in inspect.getmembers(azure_cloud) if isinstance(x[1], azure_cloud.Cloud)]
            matched_clouds = [x for x in all_clouds if x.name == cloud_name]
        except Exception as ec:
            raise CLIError("Obtain the az login's active cloud occurred exception as {0}".format(ec))

        if len(matched_clouds) != 1:
            raise CLIError("Obtain the active cloud failed, There is no or multiple matching clouds")

        cli_credentials = {
            'credentials': AzureCliCredential(),
            'subscription_id': subscription_id,
            'cloud_environment': matched_clouds[0],
        }
        return cli_credentials

    def _get_env_credentials(self, subscription_id=None):
        env_credentials = dict()
        for attribute, env_variable in AZURE_CREDENTIAL_ENV_MAPPING.items():
            env_credentials[attribute] = os.environ.get(env_variable, None)

        if env_credentials['profile']:
            credentials = self._get_profile(env_credentials['profile'], subscription_id=subscription_id)
            return credentials

        if env_credentials.get('subscription_id') is None and not self.is_ad_resource:
            return None

        if subscription_id is not None:
            env_credentials['subscription_id'] = subscription_id

        return env_credentials

    def _get_credentials(self, auth_source=None, **params):
        # Get authentication credentials.
        self.log('Getting credentials')

        arg_credentials = dict()
        for attribute, env_variable in AZURE_CREDENTIAL_ENV_MAPPING.items():
            arg_credentials[attribute] = params.get(attribute, None)

        if auth_source == 'msi':
            self.log('Retrieving credentials from MSI')
            return self._get_msi_credentials(subscription_id=params.get('subscription_id'), client_id=params.get('client_id'),
                                             _cloud_environment=params.get('cloud_environment'))

        if auth_source == 'cli':
            try:
                self.log('Retrieving credentials from Azure CLI profile')
                cli_credentials = self._get_azure_cli_credentials(subscription_id=params.get('subscription_id'))
                return cli_credentials
            except CLIError as err:
                self.fail("Azure CLI profile cannot be loaded - {0}".format(err))

        if auth_source == 'env':
            self.log('Retrieving credentials from environment')
            env_credentials = self._get_env_credentials(subscription_id=params.get('subscription_id'))
            return env_credentials

        if auth_source == 'credential_file':
            self.log("Retrieving credentials from credential file")
            profile = params.get('profile') or 'default'
            default_credentials = self._get_profile(profile, subscription_id=params.get('subscription_id'))
            return default_credentials

        # auto, precedence: module parameters -> environment variables -> default profile in ~/.azure/credentials -> azure cli
        # try module params
        if arg_credentials['profile'] is not None:
            self.log('Retrieving credentials with profile parameter.')
            credentials = self._get_profile(arg_credentials['profile'], subscription_id=params.get('subscription_id'))
            return credentials

        if arg_credentials['client_id'] or arg_credentials['ad_user']:
            self.log('Received credentials from parameters.')
            return arg_credentials

        # try environment
        env_credentials = self._get_env_credentials(subscription_id=params.get('subscription_id'))
        if env_credentials:
            self.log('Received credentials from env.')
            return env_credentials

        # try default profile from ~./azure/credentials
        default_credentials = self._get_profile(subscription_id=params.get('subscription_id'))
        if default_credentials:
            self.log('Retrieved default profile credentials from ~/.azure/credentials.')
            return default_credentials

        try:
            self.log('Retrieving credentials from AzureCLI profile')
            cli_credentials = self._get_azure_cli_credentials(subscription_id=params.get('subscription_id'))
            return cli_credentials
        except CLIError as ce:
            self.log('Error getting AzureCLI profile credentials - {0}'.format(ce))

        return None

    def _get_cloud_from_metadata_endpoint(self, metadata_endpoint):
        _knownClouds = azure_cloud.KNOWN_CLOUDS
        return next(cloud for cloud in _knownClouds if cloud.endpoints.resource_manager.lower().rstrip('/') == metadata_endpoint.lower().rstrip('/'))

    def log(self, msg, pretty_print=False):
        pass
        # Use only during module development
        # if self.debug:
        #     log_file = open('azure_rm.log', 'a')
        #     if pretty_print:
        #         log_file.write(json.dumps(msg, indent=4, sort_keys=True))
        #     else:
        #         log_file.write(msg + u'\n')



#!/usr/bin/python
#
# Copyright 2016 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_image_info

short_description: Inspect docker images

description:
  - Provide one or more image names, and the module will inspect each, returning an array of inspection results.
  - If an image does not exist locally, it will not appear in the results. If you want to check whether an image exists locally,
    you can call the module with the image name, then check whether the result list is empty (image does not exist) or has
    one element (the image exists locally).
  - The module will not attempt to pull images from registries. Use M(community.docker.docker_image) with O(community.docker.docker_image#module:source=pull)
    to ensure an image is pulled.
notes:
  - This module was called C(docker_image_facts) before Ansible 2.8. The usage did not change.
extends_documentation_fragment:
  - community.docker._docker.api_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker
  - community.docker._attributes.info_module
  - community.docker._attributes.idempotent_not_modify_state

options:
  name:
    description:
      - An image name or a list of image names. Name format will be C(name[:tag]) or C(repository/name[:tag]), where C(tag)
        is optional. If a tag is not provided, V(latest) will be used. Instead of image names, also image IDs can be used.
      - If no name is provided, a list of all images will be returned.
    type: list
    elements: str

requirements:
  - "Docker API >= 1.25"

author:
  - Chris Houseknecht (@chouseknecht)
"""

EXAMPLES = r"""
---
- name: Inspect a single image
  community.docker.docker_image_info:
    name: pacur/centos-7

- name: Inspect multiple images
  community.docker.docker_image_info:
    name:
      - pacur/centos-7
      - sinatra
  register: result

- name: Make sure that both images pacur/centos-7 and sinatra exist locally
  ansible.builtin.assert:
    that:
      - result.images | length == 2
"""

RETURN = r"""
images:
  description:
    - Inspection results for the selected images.
    - The list only contains inspection results of images existing locally.
  returned: always
  type: list
  elements: dict
  sample: [
    {
      "Architecture": "amd64",
      "Author": "",
      "Comment": "",
      "Config": {
        "AttachStderr": false,
        "AttachStdin": false,
        "AttachStdout": false,
        "Cmd": ["/etc/docker/registry/config.yml"],
        "Domainname": "",
        "Entrypoint": ["/bin/registry"],
        "Env": ["PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"],
        "ExposedPorts": {"5000/tcp": {}},
        "Hostname": "e5c68db50333",
        "Image": "c72dce2618dc8f7b794d2b2c2b1e64e0205ead5befc294f8111da23bd6a2c799",
        "Labels": {},
        "OnBuild": [],
        "OpenStdin": false,
        "StdinOnce": false,
        "Tty": false,
        "User": "",
        "Volumes": {"/var/lib/registry": {}},
        "WorkingDir": "",
      },
      "Container": "e83a452b8fb89d78a25a6739457050131ca5c863629a47639530d9ad2008d610",
      "ContainerConfig": {
        "AttachStderr": false,
        "AttachStdin": false,
        "AttachStdout": false,
        "Cmd": ["/bin/sh", "-c", '#(nop) CMD ["/etc/docker/registry/config.yml"]'],
        "Domainname": "",
        "Entrypoint": ["/bin/registry"],
        "Env": ["PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"],
        "ExposedPorts": {"5000/tcp": {}},
        "Hostname": "e5c68db50333",
        "Image": "c72dce2618dc8f7b794d2b2c2b1e64e0205ead5befc294f8111da23bd6a2c799",
        "Labels": {},
        "OnBuild": [],
        "OpenStdin": false,
        "StdinOnce": false,
        "Tty": false,
        "User": "",
        "Volumes": {"/var/lib/registry": {}},
        "WorkingDir": "",
      },
      "Created": "2016-03-08T21:08:15.399680378Z",
      "DockerVersion": "1.9.1",
      "GraphDriver": {
        "Data": null,
        "Name": "aufs",
      },
      "Id": "53773d8552f07b730f3e19979e32499519807d67b344141d965463a950a66e08",
      "Name": "registry:2",
      "Os": "linux",
      "Parent": "f0b1f729f784b755e7bf9c8c2e65d8a0a35a533769c2588f02895f6781ac0805",
      "RepoDigests": [],
      "RepoTags": ["registry:2"],
      "Size": 0,
      "VirtualSize": 165808884,
    },
  ]
"""

import traceback
import typing as t

from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    DockerException,
    NotFound,
)
from ansible_collections.community.docker.plugins.module_utils._api.utils.utils import (
    parse_repository_tag,
)
from ansible_collections.community.docker.plugins.module_utils._common_api import (
    AnsibleDockerClient,
    RequestException,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DockerBaseClass,
    is_image_name_id,
)


class ImageManager(DockerBaseClass):
    def __init__(self, client: AnsibleDockerClient, results: dict[str, t.Any]) -> None:
        super().__init__()

        self.client = client
        self.results = results
        self.name = self.client.module.params.get("name")
        self.log(f"Gathering facts for images: {self.name}")

        if self.name:
            self.results["images"] = self.get_facts()
        else:
            self.results["images"] = self.get_all_images()

    def fail(self, msg: str) -> t.NoReturn:
        self.client.fail(msg)

    def get_facts(self) -> list[dict[str, t.Any]]:
        """
        Lookup and inspect each image name found in the names parameter.

        :returns array of image dictionaries
        """

        results = []

        names = self.name
        if not isinstance(names, list):
            names = [names]

        for name in names:
            if is_image_name_id(name):
                self.log(f"Fetching image {name} (ID)")
                image = self.client.find_image_by_id(name, accept_missing_image=True)
            else:
                repository, tag = parse_repository_tag(name)
                if not tag:
                    tag = "latest"
                self.log(f"Fetching image {repository}:{tag}")
                image = self.client.find_image(name=repository, tag=tag)
            if image:
                results.append(image)
        return results

    def get_all_images(self) -> list[dict[str, t.Any]]:
        results = []
        params = {
            "only_ids": 0,
            "all": 0,
        }
        images = self.client.get_json("/images/json", params=params)
        for image in images:
            try:
                inspection = self.client.get_json("/images/{0}/json", image["Id"])
            except NotFound:
                inspection = None
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error inspecting image {image['Id']} - {exc}")
            results.append(inspection)
        return results


def main() -> None:
    argument_spec = {
        "name": {"type": "list", "elements": "str"},
    }

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    try:
        results = {"changed": False, "images": []}

        ImageManager(client, results)
        client.module.exit_json(**results)
    except DockerException as e:
        client.fail(
            f"An unexpected Docker error occurred: {e}",
            exception=traceback.format_exc(),
        )
    except RequestException as e:
        client.fail(
            f"An unexpected requests error occurred when trying to talk to the Docker daemon: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage CheckPoint Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
---
module: cp_mgmt_set_content_awareness_advanced_settings
short_description: Edit Content Awareness Blades' Settings.
description:
  - Edit Content Awareness Blades' Settings.
  - All operations are performed over Web Services API.
  - Available from R82 management version.
version_added: "6.0.0"
author: "Eden Brillant (@chkp-edenbr)"
options:
  internal_error_fail_mode:
    description:
      - In case of internal system error, allow or block all connections.
    type: str
    choices: ['allow connections', 'block connections']
  supported_services:
    description:
      - Specify the services that Content Awareness inspects.
    type: list
    elements: str
  httpi_non_standard_ports:
    description:
      - Servers usually send HTTP traffic on TCP port 80. Some servers send HTTP traffic on other ports also. By default, this setting is enabled and
        Content Awareness inspects HTTP traffic on non-standard ports. You can disable this setting and configure Content Awareness to inspect HTTP traffic
        only on port 80.
    type: bool
  inspect_archives:
    description:
      - Examine the content of archive files. For example, files with the extension .zip, .gz, .tgz, .tar.Z, .tar, .lzma, .tlz, 7z, .rar.
    type: bool
extends_documentation_fragment: check_point.mgmt.checkpoint_commands
"""

EXAMPLES = """
- name: set-content-awareness-advanced-settings
  cp_mgmt_set_content_awareness_advanced_settings:
    httpi_non_standard_ports: 'false'
    inspect_archives: 'false'
    internal_error_fail_mode: block connections
    supported_services:
      - Squid_NTLM
"""

RETURN = """
cp_mgmt_set_content_awareness_advanced_settings:
  description: The checkpoint set-content-awareness-advanced-settings output.
  returned: always.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import checkpoint_argument_spec_for_commands, api_command


def main():
    argument_spec = dict(
        internal_error_fail_mode=dict(type='str', choices=['allow connections', 'block connections']),
        supported_services=dict(type='list', elements='str'),
        httpi_non_standard_ports=dict(type='bool'),
        inspect_archives=dict(type='bool')
    )
    argument_spec.update(checkpoint_argument_spec_for_commands)

    module = AnsibleModule(argument_spec=argument_spec)

    command = "set-content-awareness-advanced-settings"

    result = api_command(module, command)
    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2023 Kevin Breit (@kbreit) <kevin.breit@kevinbreit.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["deprecated"],
    "supported_by": "community",
}

DOCUMENTATION = r"""
author:
  - Kevin Breit (@kbreit)
deprecated:
  alternative: cisco.meraki.networks_settings
  removed_in: 3.0.0
  why: Updated modules released with increased functionality
description:
  - Allows for management of settings of networks within Meraki.
extends_documentation_fragment: cisco.meraki.meraki
module: meraki_network_settings
options:
  local_status_page:
    description:
      - Configuration stanza of the local status page.
    suboptions:
      authentication:
        description:
          - Local status page authentication settings.
        suboptions:
          enabled:
            description:
              - Set whether local status page authentication is enabled.
            type: bool
          password:
            description:
              - Set password on local status page.
            type: str
        type: dict
    type: dict
  local_status_page_enabled:
    description: '- Enables the local device status pages (U[my.meraki.com](my.meraki.com),
      U[ap.meraki.com](ap.meraki.com), U[switch.meraki.com](switch.meraki.com), U[wired.meraki.com](wired.meraki.com)).
      - Only can be specified on its own or with C(remote_status_page_enabled). '
    type: bool
  net_id:
    description:
      - ID number of a network.
    type: str
  net_name:
    aliases:
      - name
      - network
    description:
      - Name of a network.
    type: str
  remote_status_page_enabled:
    description:
      - Enables access to the device status page (U(http://device LAN IP)).
      - Can only be set if C(local_status_page_enabled:) is set to C(yes).
      - Only can be specified on its own or with C(local_status_page_enabled).
    type: bool
  secure_port:
    description:
      - Configuration of SecureConnect options applied to the network.
    suboptions:
      enabled:
        description:
          - Set whether SecureConnect is enabled on the network.
        type: bool
    type: dict
  state:
    choices:
      - present
      - query
    default: query
    description:
      - Create or modify an organization.
    type: str
short_description: Manage the settings of networks in the Meraki cloud
"""

EXAMPLES = r"""
- name: Get network settings
  cisco.meraki.meraki_network_settings:
    auth_key: '{{ auth_key }}'
    state: query
    org_name: '{{ test_org_name }}'
    net_name: NetworkSettingsTestNet
  delegate_to: localhost
- name: Update network settings
  cisco.meraki.meraki_network_settings:
    auth_key: '{{ auth_key }}'
    state: present
    org_name: '{{ test_org_name }}'
    net_name: NetworkSettingsTestNet
    local_status_page_enabled: false
  delegate_to: localhost
- name: Enable password on local page
  cisco.meraki.meraki_network_settings:
    auth_key: '{{ auth_key }}'
    state: present
    org_name: '{{ test_org_name }}'
    net_name: NetworkSettingsTestNet
    local_status_page_enabled: true
    local_status_page:
      authentication:
        enabled: true
        password: abc123
  delegate_to: localhost
"""

RETURN = r"""
data:
    description: Information about the created or manipulated object.
    returned: info
    type: complex
    contains:
      local_status_page_enabled:
        description: States whether U(my.meraki.com) and other device portals should be enabled.
        returned: success
        type: bool
        sample: true
      remote_status_page_enabled:
        description: Enables access to the device status page.
        returned: success
        type: bool
        sample: true
      expire_data_older_than:
        description: The number of days, weeks, or months in Epoch time to expire the data before
        returned: success
        type: int
        sample: 1234
      fips:
        description: A hash of FIPS options applied to the Network.
        returned: success
        type: complex
        contains:
          enabled:
            description: Enables/disables FIPS on the network.
            returned: success
            type: bool
            sample: true
      local_status_page:
        description: A hash of Local Status Page(s) authentication options applied to the Network.
        returned: success
        type: complex
        contains:
          authentication:
            description: A hash of Local Status Pages' authentication options applied to the Network.
            type: complex
            contains:
              username:
                description: The username used for Local Status Pages.
                type: str
                returned: success
                sample: admin
              enabled:
                description: Enables/Disables the authenticaiton on Local Status Pages.
                type: bool
                returned: success
            sample: true
      secure_port:
        description: A hash of SecureConnect options applied to the Network.
        type: complex
        contains:
          enabled:
            description: Enables/disables SecureConnect on the network.
            type: bool
            returned: success
            sample: true
      named_vlans:
        description: A hash of Named VLANs options applied to the Network.
        type: complex
        contains:
          enabled:
            description: Enables/disables Named VLANs on the network.
            type: bool
            returned: success
            sample: true
"""

from ansible.module_utils.basic import AnsibleModule, json
from ansible_collections.cisco.meraki.plugins.module_utils.network.meraki.meraki import (
    MerakiModule,
    meraki_argument_spec,
)


def is_net_valid(data, net_name=None, net_id=None):
    if net_name is None and net_id is None:
        return False
    for n in data:
        if net_name:
            if n["name"] == net_name:
                return True
        elif net_id:
            if n["id"] == net_id:
                return True
    return False


def get_network_settings(meraki, net_id):
    path = meraki.construct_path("get_settings", net_id=net_id)
    response = meraki.request(path, method="GET")
    return response


def construct_payload(params):
    payload = dict()
    if params["local_status_page_enabled"] is not None:
        payload["localStatusPageEnabled"] = params["local_status_page_enabled"]
    if params["remote_status_page_enabled"] is not None:
        payload["remoteStatusPageEnabled"] = params["remote_status_page_enabled"]
    if params["local_status_page"] is not None:
        payload["localStatusPage"] = dict()
        if params["local_status_page"]["authentication"] is not None:
            payload["localStatusPage"]["authentication"] = {}
            if params["local_status_page"]["authentication"]["enabled"] is not None:
                payload["localStatusPage"]["authentication"]["enabled"] = params["local_status_page"]["authentication"]["enabled"]
            if params["local_status_page"]["authentication"]["password"] is not None:
                payload["localStatusPage"]["authentication"]["password"] = params["local_status_page"]["authentication"]["password"]
    if params["secure_port"] is not None:
        payload["securePort"] = dict()
        if params["secure_port"]["enabled"] is not None:
            payload["securePort"]["enabled"] = params["secure_port"]["enabled"]
    return payload


def main():

    # define the available arguments/parameters that a user can pass to
    # the module

    auth_args = dict(
        enabled=dict(type="bool"),
        password=dict(type="str", no_log=True),
    )

    local_status_page_args = dict(
        authentication=dict(type="dict", default=None, options=auth_args),
    )

    secure_port_args = dict(
        enabled=dict(type="bool"),
    )

    argument_spec = meraki_argument_spec()
    argument_spec.update(
        state=dict(type="str", choices=["query", "present"], default="query"),
        net_name=dict(type="str", aliases=["name", "network"]),
        net_id=dict(type="str"),
        local_status_page_enabled=dict(type="bool"),
        remote_status_page_enabled=dict(type="bool"),
        local_status_page=dict(type="dict", default=None, options=local_status_page_args),
        secure_port=dict(type="dict", default=None, options=secure_port_args)
    )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    meraki = MerakiModule(module, function="network")
    module.params["follow_redirects"] = "all"
    payload = None

    update_settings_urls = {"network": "/networks/{net_id}/settings"}
    get_settings_urls = {"network": "/networks/{net_id}/settings"}
    meraki.url_catalog["update_settings"] = update_settings_urls
    meraki.url_catalog["get_settings"] = get_settings_urls

    if not meraki.params["org_name"] and not meraki.params["org_id"]:
        meraki.fail_json(msg="org_name or org_id parameters are required")
    if meraki.params["state"] != "query":
        if not meraki.params["net_name"] and not meraki.params["net_id"]:
            meraki.fail_json(
                msg="net_name or net_id is required for present or absent states"
            )
    if meraki.params["net_name"] and meraki.params["net_id"]:
        meraki.fail_json(msg="net_name and net_id are mutually exclusive")
    if (
        meraki.params["local_status_page_enabled"] is False
        and meraki.params["remote_status_page_enabled"] is True
    ):
        meraki.fail_json(
            msg="local_status_page_enabled must be true when setting remote_status_page_enabled"
        )

    # manipulate or modify the state as needed (this is going to be the
    # part where your module will do what it needs to do)

    org_id = meraki.params["org_id"]
    net_id = meraki.params["net_id"]
    if not org_id:
        org_id = meraki.get_org_id(meraki.params["org_name"])
    if net_id is None:
        nets = meraki.get_nets(org_id=org_id)
        net_id = meraki.get_net_id(org_id, meraki.params["net_name"], data=nets)

    if meraki.params["state"] == "query":
        path = meraki.construct_path("get_settings", net_id=net_id)
        meraki.result["data"] = meraki.request(path, method="GET")
        meraki.exit_json(**meraki.result)
    elif meraki.params["state"] == "present":
        path = meraki.construct_path("get_settings", net_id=net_id)
        current = meraki.request(path, method="GET")
        payload = construct_payload(meraki.params)
        if meraki.is_update_required(current, payload, optional_ignore=["password"]):
            if meraki.check_mode is True:
                try:
                    del payload["local_status_page"]["authentication"]["password"]
                except KeyError:
                    pass
                meraki.result["data"] = payload
                meraki.result["changed"] = True
                meraki.exit_json(**meraki.result)
            path = meraki.construct_path("update_settings", net_id=net_id)
            response = meraki.request(path, method="PUT", payload=json.dumps(payload))
            if meraki.status == 200:
                meraki.result["changed"] = True
                meraki.result["data"] = response
                meraki.exit_json(**meraki.result)
        meraki.result["data"] = current
    # in the event of a successful module execution, you will want to
    # simple AnsibleModule.exit_json(), passing the key/value results
    meraki.exit_json(**meraki.result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage CheckPoint Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
---
module: cp_mgmt_best_practice_facts
short_description: Get best-practice objects facts on Checkpoint over Web Services API
description:
  - Get best-practice objects facts on Checkpoint devices.
  - All operations are performed over Web Services API.
  - This module handles both operations, get a specific object and get several objects,
    For getting a specific object use the parameter 'name' or 'best_practice_id'.
  - Available from R82.10 management version.
version_added: "6.7.0"
author: "Eden Brillant (@chkp-edenbr)"
options:
  best_practice_id:
    description:
      - Best Practice ID.
    type: str
  name:
    description:
      - Best Practice Name.
        This parameter is relevant only for getting a specific object.
    type: str
  show_regulations:
    description:
      - Show the applicable regulations of the Best Practice.
    type: bool
  details_level:
    description:
      - The level of detail for some of the fields in the response can vary from showing only the UID value of the object to a fully detailed
        representation of the object.
    type: str
    choices: ['uid', 'standard', 'full']
  blade:
    description:
      - Returns all the relevant Best Practices of the selected Software Blades. When empty will return all the Best Practices.
    type: list
    elements: str
    choices: ['firewall', 'gaia-os', 'anti-bot', 'anti-spam-and-mail', 'anti-virus', 'application-control', 'data-loss-prevention',
             'identity-awareness', 'ips', 'ipsec-vpn', 'mobile-access', 'threat-emulation', 'url-filtering', 'threat-prevention']
  limit:
    description:
      - The maximal number of returned results.
        This parameter is relevant only for getting few objects.
      - Valid values are between 1 and 500.
    type: int
  offset:
    description:
      - Number of the results to initially skip.
        This parameter is relevant only for getting few objects.
    type: int
  status:
    description:
      - Returns all the relevant best practices with the selected statuses. When empty will return all best practices.
    type: list
    elements: str
    choices: ['secure', 'good', 'medium', 'poor', 'n/a']
  gateway_name:
    description:
      - Returns all the relevant Best Practices of the selected Security Gateway object.
    type: str
  defined_by:
    description:
      - Returns all the relevant Best Practices of the selected type.
    type: str
    choices: ['user-defined', 'system-defined', 'any']
  show_only_local_domain:
    description:
      - Indicates whether the query should return only objects from the current local domain. This parameter is only valid for local domain.
    type: bool
extends_documentation_fragment: check_point.mgmt.checkpoint_facts
"""

EXAMPLES = """
- name: show-best-practice
  cp_mgmt_best_practice_facts:
    best_practice_id: FW183

- name: show-best-practices
  cp_mgmt_best_practice_facts:
    limit: '5'
"""

RETURN = """
ansible_facts:
  description: The checkpoint object facts.
  returned: always.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import checkpoint_argument_spec_for_facts, api_call_facts


def main():
    argument_spec = dict(
        best_practice_id=dict(type='str'),
        name=dict(type='str'),
        show_regulations=dict(type='bool'),
        details_level=dict(type='str', choices=['uid', 'standard', 'full']),
        blade=dict(type='list', elements='str', choices=['firewall', 'gaia-os', 'anti-bot', 'anti-spam-and-mail', 'anti-virus',
                                                         'application-control', 'data-loss-prevention', 'identity-awareness', 'ips',
                                                         'ipsec-vpn', 'mobile-access', 'threat-emulation', 'url-filtering', 'threat-prevention']),
        limit=dict(type='int'),
        offset=dict(type='int'),
        status=dict(type='list', elements='str', choices=['secure', 'good', 'medium', 'poor', 'n/a']),
        gateway_name=dict(type='str'),
        defined_by=dict(type='str', choices=['user-defined', 'system-defined', 'any']),
        show_only_local_domain=dict(type='bool')
    )
    argument_spec.update(checkpoint_argument_spec_for_facts)

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    api_call_object = "best-practice"
    api_call_object_plural_version = "best-practices"

    result = api_call_facts(module, api_call_object, api_call_object_plural_version)
    module.exit_json(ansible_facts=result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_config_templates_switch_profiles_ports_info
short_description: Information module for organizations _config _templates _switch
  _profiles _ports
description:
  - Get all organizations _config _templates _switch _profiles _ports.
  - Get organizations _config _templates _switch _profiles _ports by id.
  - Return a switch template port.
  - Return all the ports of a switch template.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  configTemplateId:
    description:
      - ConfigTemplateId path parameter. Config template ID.
    type: str
  profileId:
    description:
      - ProfileId path parameter. Profile ID.
    type: str
  portId:
    description:
      - PortId path parameter. Port ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for switch getOrganizationConfigTemplateSwitchProfilePort
    description: Complete reference of the getOrganizationConfigTemplateSwitchProfilePort
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-config-template-switch-profile-port
  - name: Cisco Meraki documentation for switch getOrganizationConfigTemplateSwitchProfilePorts
    description: Complete reference of the getOrganizationConfigTemplateSwitchProfilePorts
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-config-template-switch-profile-ports
notes:
  - SDK Method used are
    switch.Switch.get_organization_config_template_switch_profile_port,
    switch.Switch.get_organization_config_template_switch_profile_ports,
  - Paths used are
    get /organizations/{organizationId}/configTemplates/{configTemplateId}/switch/profiles/{profileId}/ports,
    get /organizations/{organizationId}/configTemplates/{configTemplateId}/switch/profiles/{profileId}/ports/{portId},
"""

EXAMPLES = r"""
- name: Get all organizations _config _templates _switch _profiles _ports
  cisco.meraki.organizations_config_templates_switch_profiles_ports_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    organizationId: string
    configTemplateId: string
    profileId: string
  register: result
- name: Get organizations _config _templates _switch _profiles _ports by id
  cisco.meraki.organizations_config_templates_switch_profiles_ports_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    organizationId: string
    configTemplateId: string
    profileId: string
    portId: string
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "portId": "string",
      "name": "string",
      "tags": [
        "string"
      ],
      "enabled": true,
      "poeEnabled": true,
      "type": "string",
      "vlan": 0,
      "voiceVlan": 0,
      "allowedVlans": "string",
      "isolationEnabled": true,
      "rstpEnabled": true,
      "stpGuard": "string",
      "stpPortFastTrunk": true,
      "linkNegotiation": "string",
      "linkNegotiationCapabilities": [
        "string"
      ],
      "portScheduleId": "string",
      "schedule": {
        "id": "string",
        "name": "string"
      },
      "udld": "string",
      "accessPolicyType": "string",
      "accessPolicyNumber": 0,
      "macAllowList": [
        "string"
      ],
      "macWhitelistLimit": 0,
      "stickyMacAllowList": [
        "string"
      ],
      "stickyMacAllowListLimit": 0,
      "stormControlEnabled": true,
      "flexibleStackingEnabled": true,
      "daiTrusted": true,
      "profile": {
        "enabled": true,
        "id": "string",
        "iname": "string"
      },
      "module": {
        "model": "string",
        "serial": "string",
        "slot": 0
      },
      "mirror": {
        "mode": "string"
      },
      "dot3az": {
        "enabled": true
      },
      "highSpeed": {
        "enabled": true
      }
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2023, Tim Cragg (@timcragg) <tcragg@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_file_remote_path
short_description: Manage Import/Export File Remote Paths (file:RemotePath)
description:
- Manage Import/Export File Remote Paths on Cisco ACI fabrics.
options:
  name:
    description:
    - The name of the File Remote Path.
    type: str
  description:
    description:
    - The description of the File Remote Path.
    type: str
  remote_host:
    description:
    - The hostname or IP Address of the remote host.
    type: str
  remote_port:
    description:
    - The port to access the remote host.
    type: int
  remote_protocol:
    description:
    - The protocol to use to connect to the remote host.
    type: str
    choices: [ ftp, scp, sftp ]
  auth_type:
    description:
    - The authentication type for the remote host.
    - Cannot be set to ssh_key if protocol is ftp.
    type: str
    choices: [ password, ssh_key ]
  remote_user:
    description:
    - The username to access the remote host.
    type: str
  remote_password:
    description:
    - The password to access the remote host.
    - Only used if auth_type is password.
    - Providing this option will always result in a change because it is a secure property that cannot be retrieved from APIC.
    type: str
  remote_ssh_key:
    description:
    - The private SSH key used to access the remote host.
    - Only used if auth_type is ssh_key.
    - Providing this option will always result in a change because it is a secure property that cannot be retrieved from APIC.
    type: str
    aliases: [ remote_key ]
  remote_ssh_passphrase:
    description:
    - The Pass phrase used to decode private_key.
    - Only used if auth_type is ssh_key.
    - Providing this option will always result in a change because it is a secure property that cannot be retrieved from APIC.
    type: str
    aliases: [ passphrase ]
  remote_path:
    description:
    - The path on which the data will reside on the remote host.
    type: str
  management_epg:
    description:
    - The management EPG to connect to the remote host on.
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(file:RemotePath).
  link: https://developer.cisco.com/docs/apic-mim-ref/

author:
- Tim Cragg (@timcragg)
"""

EXAMPLES = r"""
- name: Add a Remote Path
  cisco.aci.aci_file_remote_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: ans_remote_path
    remote_host: test.example.com
    remote_port: 22
    remote_protocol: scp
    remote_user: test_user
    auth_type: password
    remote_password: test_pass
    remote_path: /tmp
    state: present
  delegate_to: localhost

- name: Query a Remote Path
  cisco.aci.aci_file_remote_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: ans_remote_path
    state: query
  delegate_to: localhost

- name: Query all Remote Paths
  cisco.aci.aci_file_remote_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost

- name: Remove a Remote Path
  cisco.aci.aci_file_remote_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: ans_remote_path
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        name=dict(type="str"),
        description=dict(type="str"),
        remote_host=dict(type="str"),
        remote_port=dict(type="int"),
        remote_protocol=dict(type="str", choices=["ftp", "scp", "sftp"]),
        remote_path=dict(type="str"),
        auth_type=dict(type="str", choices=["password", "ssh_key"]),
        remote_user=dict(type="str"),
        remote_password=dict(type="str", no_log=True),
        remote_ssh_key=dict(type="str", no_log=True, aliases=["remote_key"]),
        remote_ssh_passphrase=dict(type="str", no_log=True, aliases=["passphrase"]),
        management_epg=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["name", "auth_type"]],
            ["state", "absent", ["name"]],
        ],
        required_together=[["remote_host", "remote_port"]],
    )

    name = module.params.get("name")
    description = module.params.get("description")
    remote_host = module.params.get("remote_host")
    remote_port = module.params.get("remote_port")
    remote_protocol = module.params.get("remote_protocol")
    remote_path = module.params.get("remote_path")
    auth_type = module.params.get("auth_type")
    remote_user = module.params.get("remote_user")
    remote_password = module.params.get("remote_password")
    remote_key = module.params.get("remote_ssh_key")
    passphrase = module.params.get("remote_ssh_passphrase")
    management_epg = module.params.get("management_epg")
    state = module.params.get("state")

    aci = ACIModule(module)

    if auth_type == "password":
        if remote_key is not None:
            aci.fail_json(msg="remote_key cannot be set if auth_type is password")
        if passphrase is not None:
            aci.fail_json(msg="passphrase cannot be set if auth_type is password")
        auth = "usePassword"
    elif auth_type == "ssh_key":
        if remote_password is not None:
            aci.fail_json(msg="remote_password cannot be set if auth_type is ssh_key")
        auth = "useSshKeyContents"
    else:
        auth = None

    aci.construct_url(
        root_class=dict(
            aci_class="fileRemotePath",
            aci_rn="fabric/path-{0}".format(name),
            module_object=name,
            target_filter={"name": name},
        ),
        child_classes=["fileRsARemoteHostToEpg"],
    )
    aci.get_existing()

    if state == "present":
        child_configs = []
        if management_epg is not None:
            child_configs.append(
                dict(
                    fileRsARemoteHostToEpg=dict(
                        attributes=dict(tDn=("uni/tn-mgmt/mgmtp-default/{0}".format(management_epg))),
                    )
                )
            )
        aci.payload(
            aci_class="fileRemotePath",
            class_config=dict(
                name=name,
                descr=description,
                authType=auth,
                host=remote_host,
                protocol=remote_protocol,
                remotePath=remote_path,
                remotePort=remote_port,
                userName=remote_user,
                userPasswd=remote_password,
                identityPrivateKeyContents=remote_key,
                identityPrivateKeyPassphrase=passphrase,
            ),
            child_configs=child_configs,
        )

        aci.get_diff(aci_class="fileRemotePath")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



# Copyright (c) 2017, Ansible Project
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import traceback
import typing as t

from ansible.module_utils.basic import missing_required_lib

from ansible_collections.community.general.plugins.module_utils._version import LooseVersion

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule

REQUESTS_IMP_ERR = None
try:
    import requests.exceptions  # noqa: F401, pylint: disable=unused-import

    HAS_REQUESTS = True
except ImportError:
    REQUESTS_IMP_ERR = traceback.format_exc()
    HAS_REQUESTS = False

INFLUXDB_IMP_ERR = None
try:
    from influxdb import (  # pylint: disable=unused-import
        InfluxDBClient,
        exceptions,  # noqa: F401
    )
    from influxdb import __version__ as influxdb_version

    HAS_INFLUXDB = True
except ImportError:
    INFLUXDB_IMP_ERR = traceback.format_exc()
    HAS_INFLUXDB = False


class InfluxDb:
    def __init__(self, module: AnsibleModule) -> None:
        self.module = module
        self.params = self.module.params
        self.check_lib()
        self.hostname = self.params["hostname"]
        self.port = self.params["port"]
        self.path = self.params["path"]
        self.username = self.params["username"]
        self.password = self.params["password"]
        self.database_name = self.params.get("database_name")

    def check_lib(self) -> None:
        if not HAS_REQUESTS:
            self.module.fail_json(msg=missing_required_lib("requests"), exception=REQUESTS_IMP_ERR)

        if not HAS_INFLUXDB:
            self.module.fail_json(msg=missing_required_lib("influxdb"), exception=INFLUXDB_IMP_ERR)

    @staticmethod
    def influxdb_argument_spec() -> dict[str, t.Any]:
        return dict(
            hostname=dict(type="str", default="localhost"),
            port=dict(type="int", default=8086),
            path=dict(type="str", default=""),
            username=dict(type="str", default="root", aliases=["login_username"]),
            password=dict(type="str", default="root", no_log=True, aliases=["login_password"]),
            ssl=dict(type="bool", default=False),
            validate_certs=dict(type="bool", default=True),
            timeout=dict(type="int"),
            retries=dict(type="int", default=3),
            proxies=dict(type="dict", default={}),
            use_udp=dict(type="bool", default=False),
            udp_port=dict(type="int", default=4444),
        )

    def connect_to_influxdb(self) -> InfluxDBClient:
        args = dict(
            host=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
            database=self.database_name,
            ssl=self.params["ssl"],
            verify_ssl=self.params["validate_certs"],
            timeout=self.params["timeout"],
            use_udp=self.params["use_udp"],
            udp_port=self.params["udp_port"],
            proxies=self.params["proxies"],
        )
        influxdb_api_version = LooseVersion(influxdb_version)
        if influxdb_api_version >= LooseVersion("4.1.0"):
            # retries option is added in version 4.1.0
            args.update(retries=self.params["retries"])

        if influxdb_api_version >= LooseVersion("5.1.0"):
            # path argument is added in version 5.1.0
            args.update(path=self.path)

        return InfluxDBClient(**args)
